from datetime import datetime
from typing import List, Optional, Union

from watchmen_meta.common import AuditableShaper, LastVisitShaper, UserBasedTupleService, UserBasedTupleShaper
from watchmen_model.chart import Chart
from watchmen_model.common import ConnectedSpaceId, DataPage, GraphicRect, Pageable, ParameterJoint, ReportId, \
	SubjectId, TenantId
from watchmen_model.console import Report, ReportDimension, ReportFunnel, ReportIndicator
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaJoint, \
	EntityCriteriaJointConjunction, EntityCriteriaOperator, EntityRow, EntityShaper
from watchmen_utilities import ArrayHelper


class ReportShaper(EntityShaper):
	# noinspection PyMethodMayBeStatic
	def serialize_to_dict(
			self,
			data: Optional[Union[
				dict, Chart, GraphicRect, ReportFunnel, ReportIndicator, ReportDimension, ParameterJoint
			]]
	) -> Optional[dict]:
		if data is None:
			return None
		elif isinstance(data, dict):
			return data
		else:
			return data.dict()

	def serialize(self, report: Report) -> EntityRow:
		row = {
			'report_id': report.reportId,
			'name': report.name,
			'subject_id': report.subjectId,
			'connect_id': report.connectId,
			'filters': self.serialize_to_dict(report.filters),
			'funnels': ArrayHelper(report.funnels).map(lambda x: self.serialize_to_dict(x)).to_list(),
			'indicators': ArrayHelper(report.indicators).map(lambda x: self.serialize_to_dict(x)).to_list(),
			'dimensions': ArrayHelper(report.dimensions).map(lambda x: self.serialize_to_dict(x)).to_list(),
			'description': report.description,
			'rect': self.serialize_to_dict(report.rect),
			'chart': self.serialize_to_dict(report.chart),
			'simulating': False if report.simulating is None else report.simulating,
			'simulate_data': report.simulateData,
			'simulate_thumbnail': report.simulateThumbnail
		}
		row = AuditableShaper.serialize(report, row)
		row = UserBasedTupleShaper.serialize(report, row)
		row = LastVisitShaper.serialize(report, row)
		return row

	def deserialize(self, row: EntityRow) -> Report:
		report = Report(
			reportId=row.get('report_id'),
			name=row.get('name'),
			subjectId=row.get('subject_id'),
			connectId=row.get('connect_id'),
			filters=row.get('filters'),
			funnels=row.get('funnels'),
			indicators=row.get('indicators'),
			dimensions=row.get('dimensions'),
			description=row.get('description'),
			rect=row.get('rect'),
			chart=row.get('chart'),
			simulating=row.get('simulating'),
			simulateData=row.get('simulate_data'),
			simulateThumbnail=row.get('simulate_thumbnail')
		)
		# noinspection PyTypeChecker
		report: Report = AuditableShaper.deserialize(row, report)
		# noinspection PyTypeChecker
		report: Report = UserBasedTupleShaper.deserialize(row, report)
		# noinspection PyTypeChecker
		report: Report = LastVisitShaper.deserialize(row, report)
		return report


REPORT_ENTITY_NAME = 'reports'
REPORT_ENTITY_SHAPER = ReportShaper()


class ReportService(UserBasedTupleService):
	def get_entity_name(self) -> str:
		return REPORT_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return REPORT_ENTITY_SHAPER

	def get_storable_id_column_name(self) -> str:
		return 'report_id'

	def get_storable_id(self, storable: Report) -> ReportId:
		return storable.reportId

	def set_storable_id(self, storable: Report, storable_id: ReportId) -> Report:
		storable.reportId = storable_id
		return storable

	def find_by_connect_id(self, connect_id: ConnectedSpaceId) -> List[Report]:
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='connect_id'), right=connect_id)
			]
		))

	# noinspection DuplicatedCode
	def find_page_by_text(self, text: Optional[str], tenant_id: Optional[TenantId], pageable: Pageable) -> DataPage:
		criteria = []
		if text is not None and len(text.strip()) != 0:
			criteria.append(EntityCriteriaJoint(
				conjunction=EntityCriteriaJointConjunction.OR,
				children=[
					EntityCriteriaExpression(
						left=ColumnNameLiteral(columnName='name'), operator=EntityCriteriaOperator.LIKE, right=text),
					EntityCriteriaExpression(
						left=ColumnNameLiteral(columnName='description'), operator=EntityCriteriaOperator.LIKE,
						right=text)
				]
			))
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		return self.storage.page(self.get_entity_pager(criteria=criteria, pageable=pageable))

	def delete_by_connect_id(self, connect_id: ConnectedSpaceId) -> List[Report]:
		# noinspection PyTypeChecker
		return self.storage.delete_and_pull(self.get_entity_deleter(
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='connect_id'), right=connect_id)
			]
		))

	def delete_by_subject_id(self, subject_id: SubjectId) -> List[Report]:
		# noinspection PyTypeChecker
		return self.storage.delete_and_pull(self.get_entity_deleter(
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='subject_id'), right=subject_id)
			]
		))

	def update_last_visit_time(self, report_id: ReportId) -> datetime:
		now = self.now()
		self.storage.update(self.get_entity_updater(
			criteria=[EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName=self.get_storable_id_column_name()), right=report_id)],
			update={'last_visit_time': now}
		))
		return now



from typing import List, Optional

from watchmen_meta.common import TupleService, TupleShaper
from watchmen_model.admin import Space
from watchmen_model.admin.space import SpaceFilter
from watchmen_model.common import DataPage, Pageable, SpaceId, TenantId
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaJoint, \
	EntityCriteriaJointConjunction, EntityCriteriaOperator, EntityRow, EntityShaper
from watchmen_utilities import ArrayHelper, is_not_blank


class SpaceShaper(EntityShaper):
	@staticmethod
	def serialize_filter(space_filter: SpaceFilter) -> dict:
		if isinstance(space_filter, dict):
			return space_filter
		else:
			return space_filter.dict()

	@staticmethod
	def serialize_filters(filters: Optional[List[SpaceFilter]]) -> Optional[list]:
		if filters is None:
			return None
		return ArrayHelper(filters).map(lambda x: SpaceShaper.serialize_filter(x)).to_list()

	def serialize(self, space: Space) -> EntityRow:
		return TupleShaper.serialize_tenant_based(space, {
			'space_id': space.spaceId,
			'name': space.name,
			'description': space.description,
			'topic_ids': space.topicIds,
			'group_ids': space.groupIds,
			'filters': SpaceShaper.serialize_filters(space.filters),
		})

	def deserialize(self, row: EntityRow) -> Space:
		# noinspection PyTypeChecker
		return TupleShaper.deserialize_tenant_based(row, Space(
			spaceId=row.get('space_id'),
			name=row.get('name'),
			description=row.get('description'),
			topicIds=row.get('topic_ids'),
			groupIds=row.get('group_ids'),
			filters=row.get('filters')
		))


SPACE_ENTITY_NAME = 'spaces'
SPACE_ENTITY_SHAPER = SpaceShaper()


class SpaceService(TupleService):
	def get_entity_name(self) -> str:
		return SPACE_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return SPACE_ENTITY_SHAPER

	def get_storable_id(self, storable: Space) -> SpaceId:
		return storable.spaceId

	def set_storable_id(self, storable: Space, storable_id: SpaceId) -> Space:
		storable.spaceId = storable_id
		return storable

	def get_storable_id_column_name(self) -> str:
		return 'space_id'

	# noinspection DuplicatedCode
	def find_page_by_text(self, text: Optional[str], tenant_id: Optional[TenantId], pageable: Pageable) -> DataPage:
		criteria = []
		if text is not None and len(text.strip()) != 0:
			criteria.append(EntityCriteriaJoint(
				conjunction=EntityCriteriaJointConjunction.OR,
				children=[
					EntityCriteriaExpression(
						left=ColumnNameLiteral(columnName='name'), operator=EntityCriteriaOperator.LIKE, right=text),
					EntityCriteriaExpression(
						left=ColumnNameLiteral(columnName='description'), operator=EntityCriteriaOperator.LIKE,
						right=text)
				]
			))
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		return self.storage.page(self.get_entity_pager(criteria=criteria, pageable=pageable))

	# noinspection DuplicatedCode
	def find_by_name(self, text: Optional[str], tenant_id: Optional[TenantId]) -> List[Space]:
		criteria = []
		if is_not_blank(text):
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='name'), operator=EntityCriteriaOperator.LIKE, right=text.strip()))
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria=criteria))

	def find_by_ids(self, space_ids: List[SpaceId], tenant_id: Optional[TenantId]) -> List[Space]:
		criteria = [
			EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='space_id'), operator=EntityCriteriaOperator.IN, right=space_ids)
		]
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria))

	# noinspection DuplicatedCode
	def find_all(self, tenant_id: Optional[TenantId]) -> List[Space]:
		criteria = []
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria=criteria))



from datetime import datetime
from typing import List, Optional, Union

from watchmen_meta.common import AuditableShaper, LastVisitShaper, TupleNotFoundException, UserBasedTupleService, \
	UserBasedTupleShaper
from watchmen_model.common import ConnectedSpaceId, SubjectId, TenantId, UserId
from watchmen_model.console import Subject, SubjectDataset
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaOperator, EntityRow, \
	EntityShaper


class SubjectShaper(EntityShaper):
	@staticmethod
	def serialize_dataset(dataset: Optional[Union[dict, SubjectDataset]]) -> Optional[dict]:
		if dataset is None:
			return None
		elif isinstance(dataset, dict):
			return dataset
		else:
			return dataset.dict()

	def serialize(self, subject: Subject) -> EntityRow:
		row = {
			'subject_id': subject.subjectId,
			'name': subject.name,
			'connect_id': subject.connectId,
			'auto_refresh_interval': subject.autoRefreshInterval,
			'dataset': SubjectShaper.serialize_dataset(subject.dataset)
		}
		row = AuditableShaper.serialize(subject, row)
		row = UserBasedTupleShaper.serialize(subject, row)
		row = LastVisitShaper.serialize(subject, row)
		return row

	def deserialize(self, row: EntityRow) -> Subject:
		subject = Subject(
			subjectId=row.get('subject_id'),
			name=row.get('name'),
			connectId=row.get('connect_id'),
			autoRefreshInterval=row.get('auto_refresh_interval'),
			dataset=row.get('dataset')
		)
		# noinspection PyTypeChecker
		subject: Subject = AuditableShaper.deserialize(row, subject)
		# noinspection PyTypeChecker
		subject: Subject = UserBasedTupleShaper.deserialize(row, subject)
		# noinspection PyTypeChecker
		subject: Subject = LastVisitShaper.deserialize(row, subject)
		return subject


SUBJECT_ENTITY_NAME = 'subjects'
SUBJECT_ENTITY_SHAPER = SubjectShaper()


class SubjectService(UserBasedTupleService):
	def get_entity_name(self) -> str:
		return SUBJECT_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return SUBJECT_ENTITY_SHAPER

	def get_storable_id_column_name(self) -> str:
		return 'subject_id'

	def get_storable_id(self, storable: Subject) -> SubjectId:
		return storable.subjectId

	def set_storable_id(self, storable: Subject, storable_id: SubjectId) -> Subject:
		storable.subjectId = storable_id
		return storable

	def find_by_name(self, name: str) -> Optional[Subject]:
		return self.storage.find_one(self.get_entity_finder(
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='name'), right=name)
			]
		))

	def find_by_text(self, text: Optional[str], tenant_id: Optional[TenantId]) -> List[Subject]:
		criteria = []
		if text is not None and len(text.strip()) != 0:
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='name'), operator=EntityCriteriaOperator.LIKE, right=text))
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria=criteria))

	def find_by_connect_id(self, connect_id: ConnectedSpaceId) -> List[Subject]:
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='connect_id'), right=connect_id)
			]
		))

	# noinspection DuplicatedCode
	def update_name(self, subject_id: SubjectId, name: str, user_id: UserId, tenant_id: TenantId) -> datetime:
		"""
		update name will not increase optimistic lock version
		"""
		last_modified_at = self.now()
		last_modified_by = self.principalService.get_user_id()
		updated_count = self.storage.update_only(self.get_entity_updater(
			criteria=[
				EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName=self.get_storable_id_column_name()), right=subject_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='user_id'), right=user_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
			],
			update={
				'name': name,
				'last_modified_at': last_modified_at,
				'last_modified_by': last_modified_by
			}
		))
		if updated_count == 0:
			raise TupleNotFoundException('Update 0 row might be caused by tuple not found.')
		return last_modified_at

	def update_last_visit_time(self, subject_id: SubjectId) -> datetime:
		now = self.now()
		self.storage.update(self.get_entity_updater(
			criteria=[EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName=self.get_storable_id_column_name()), right=subject_id)],
			update={'last_visit_time': now}
		))
		return now

	def delete_by_connect_id(self, connect_id: ConnectedSpaceId) -> List[Subject]:
		# noinspection PyTypeChecker
		return self.storage.delete_and_pull(self.get_entity_deleter(
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='connect_id'), right=connect_id)
			]
		))



from typing import Optional, Union

from watchmen_meta.common import UserBasedTupleService, UserBasedTupleShaper
from watchmen_model.common import ConnectedSpaceId
from watchmen_model.console import ConnectedSpaceGraphic, SubjectGraphic, TopicGraphic
from watchmen_storage import EntityRow, EntityShaper
from watchmen_utilities import ArrayHelper


class ConnectedSpaceGraphicShaper(EntityShaper):
	# noinspection PyMethodMayBeStatic
	def serialize_to_dict(
			self, data: Optional[Union[dict, TopicGraphic, SubjectGraphic]]
	) -> Optional[dict]:
		if data is None:
			return None
		elif isinstance(data, dict):
			return data
		else:
			return data.dict()

	def serialize(self, connected_space_graphic: ConnectedSpaceGraphic) -> EntityRow:
		row = {
			'connect_id': connected_space_graphic.connectId,
			'topics': ArrayHelper(connected_space_graphic.topics).map(lambda x: self.serialize_to_dict(x)).to_list(),
			'subjects': ArrayHelper(connected_space_graphic.subjects).map(lambda x: self.serialize_to_dict(x)).to_list()
		}
		row = UserBasedTupleShaper.serialize(connected_space_graphic, row)
		return row

	def deserialize(self, row: EntityRow) -> ConnectedSpaceGraphic:
		connected_space_graphic = ConnectedSpaceGraphic(
			connectId=row.get('connect_id'),
			topics=row.get('topics'),
			subjects=row.get('subjects')
		)
		# noinspection PyTypeChecker
		connected_space_graphic: ConnectedSpaceGraphic = UserBasedTupleShaper.deserialize(row, connected_space_graphic)
		return connected_space_graphic


CONNECTED_SPACE_GRAPHIC_ENTITY_NAME = 'connected_space_graphics'
CONNECTED_SPACE_GRAPHIC_ENTITY_SHAPER = ConnectedSpaceGraphicShaper()


class ConnectedSpaceGraphicService(UserBasedTupleService):
	def get_entity_name(self) -> str:
		return CONNECTED_SPACE_GRAPHIC_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return CONNECTED_SPACE_GRAPHIC_ENTITY_SHAPER

	def get_storable_id_column_name(self) -> str:
		return 'connect_id'

	def get_storable_id(self, storable: ConnectedSpaceGraphic) -> ConnectedSpaceId:
		return storable.connectId

	def set_storable_id(self, storable: ConnectedSpaceGraphic, storable_id: ConnectedSpaceId) -> ConnectedSpaceGraphic:
		storable.connectId = storable_id
		return storable



from typing import List, Optional

from watchmen_meta.common import TupleService, TupleShaper
from watchmen_model.common import DataPage, DataSourceId, Pageable, TenantId
from watchmen_model.system import DataSource, DataSourceParam
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaOperator, EntityRow, \
	EntityShaper
from watchmen_utilities import ArrayHelper


class DataSourceShaper(EntityShaper):
	@staticmethod
	def serialize_param(param: Optional[DataSourceParam]) -> Optional[dict]:
		if param is None:
			return None
		if isinstance(param, dict):
			return param
		else:
			return param.to_dict()

	@staticmethod
	def serialize_params(params: Optional[List[DataSourceParam]]) -> Optional[list]:
		if params is None:
			return None
		return ArrayHelper(params).map(lambda x: DataSourceShaper.serialize_param(x)).to_list()

	def serialize(self, data_source: DataSource) -> EntityRow:
		return TupleShaper.serialize_tenant_based(data_source, {
			'data_source_id': data_source.dataSourceId,
			'data_source_code': data_source.dataSourceCode,
			'data_source_type': data_source.dataSourceType,
			'host': data_source.host,
			'port': data_source.port,
			'username': data_source.username,
			'password': data_source.password,
			'name': data_source.name,
			'url': data_source.url,
			'params': DataSourceShaper.serialize_params(data_source.params)
		})

	def deserialize(self, row: EntityRow) -> DataSource:
		# noinspection PyTypeChecker
		return TupleShaper.deserialize_tenant_based(row, DataSource(
			dataSourceId=row.get('data_source_id'),
			dataSourceCode=row.get('data_source_code'),
			dataSourceType=row.get('data_source_type'),
			host=row.get('host'),
			port=row.get('port'),
			username=row.get('username'),
			password=row.get('password'),
			name=row.get('name'),
			url=row.get('url'),
			params=row.get('params')
		))


DATA_SOURCE_ENTITY_NAME = 'data_sources'
DATA_SOURCE_ENTITY_SHAPER = DataSourceShaper()


class DataSourceService(TupleService):
	def get_entity_name(self) -> str:
		return DATA_SOURCE_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return DATA_SOURCE_ENTITY_SHAPER

	def get_storable_id(self, storable: DataSource) -> DataSourceId:
		return storable.dataSourceId

	def set_storable_id(self, storable: DataSource, storable_id: DataSourceId) -> DataSource:
		storable.dataSourceId = storable_id
		return storable

	def get_storable_id_column_name(self) -> str:
		return 'data_source_id'

	# noinspection DuplicatedCode
	def find_by_text(
			self, text: Optional[str], tenant_id: Optional[TenantId], pageable: Pageable) -> DataPage:
		criteria = []
		if text is not None and len(text.strip()) != 0:
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='name'), operator=EntityCriteriaOperator.LIKE, right=text))
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		return self.storage.page(self.get_entity_pager(criteria, pageable))

	def find_all(self, tenant_id: Optional[TenantId]) -> List[DataSource]:
		criteria = []
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria))



from .data_source_service import DataSourceService
from .external_writer_service import ExternalWriterService
from .key_store_service import KeyStoreService
from .pat_service import PatService
from .tenant_service import TenantService



from typing import List, Optional

from watchmen_meta.common import TupleService, TupleShaper
from watchmen_model.admin import TopicSnapshotFrequency, TopicSnapshotScheduler, TopicSnapshotSchedulerId
from watchmen_model.common import DataPage, Pageable, ParameterJoint, TenantId, TopicId
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaOperator, EntityRow, \
	EntityShaper
from watchmen_utilities import is_not_blank


class TopicSnapshotSchedulerShaper(EntityShaper):
	@staticmethod
	def serialize_filter(a_filter: ParameterJoint) -> Optional[dict]:
		if a_filter is None:
			return None
		elif isinstance(a_filter, dict):
			return a_filter
		else:
			return a_filter.dict()

	def serialize(self, scheduler: TopicSnapshotScheduler) -> EntityRow:
		return TupleShaper.serialize_tenant_based(scheduler, {
			'scheduler_id': scheduler.schedulerId,
			'topic_id': scheduler.topicId,
			'target_topic_name': scheduler.targetTopicName,
			'target_topic_id': scheduler.targetTopicId,
			'pipeline_id': scheduler.pipelineId,
			'frequency': scheduler.frequency,
			'filter': TopicSnapshotSchedulerShaper.serialize_filter(scheduler.filter),
			'weekday': scheduler.weekday,
			'day': scheduler.day,
			'hour': scheduler.hour,
			'minute': scheduler.minute,
			'enabled': True if scheduler.enabled is None else scheduler.enabled
		})

	def deserialize(self, row: EntityRow) -> TopicSnapshotScheduler:
		# noinspection PyTypeChecker
		return TupleShaper.deserialize_tenant_based(row, TopicSnapshotScheduler(
			schedulerId=row.get('scheduler_id'),
			topicId=row.get('topic_id'),
			targetTopicName=row.get('target_topic_name'),
			targetTopicId=row.get('target_topic_id'),
			pipelineId=row.get('pipeline_id'),
			frequency=row.get('frequency'),
			filter=row.get('filter'),
			weekday=row.get('weekday'),
			day=row.get('day'),
			hour=row.get('hour'),
			minute=row.get('minute'),
			enabled=row.get('enabled')
		))


TOPIC_SNAPSHOT_SCHEDULER_ENTITY_NAME = 'snapshot_schedulers'
TOPIC_SNAPSHOT_SCHEDULER_ENTITY_SHAPER = TopicSnapshotSchedulerShaper()


class TopicSnapshotSchedulerService(TupleService):
	def get_entity_name(self) -> str:
		return TOPIC_SNAPSHOT_SCHEDULER_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return TOPIC_SNAPSHOT_SCHEDULER_ENTITY_SHAPER

	def get_storable_id(self, storable: TopicSnapshotScheduler) -> TopicSnapshotSchedulerId:
		return storable.schedulerId

	def set_storable_id(
			self, storable: TopicSnapshotScheduler, storable_id: TopicSnapshotSchedulerId) -> TopicSnapshotScheduler:
		storable.schedulerId = storable_id
		return storable

	def get_storable_id_column_name(self) -> str:
		return 'scheduler_id'

	def find_page_by_topic_and_frequency(
			self, topic_id: Optional[TopicId], frequency: Optional[List[TopicSnapshotFrequency]],
			tenant_id: Optional[TenantId],
			pageable: Pageable
	) -> DataPage:
		criteria = []
		if is_not_blank(topic_id):
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='topic_id'), operator=EntityCriteriaOperator.EQUALS, right=topic_id))
		if frequency is not None and len(frequency) != 0:
			if len(frequency) == 1:
				criteria.append(EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName='frequency'), operator=EntityCriteriaOperator.EQUALS,
					right=frequency[0]))
			else:
				criteria.append(EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName='frequency'), operator=EntityCriteriaOperator.IN,
					right=frequency))
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		return self.storage.page(self.get_entity_pager(criteria=criteria, pageable=pageable))

	def find_by_topic(self, topic_id: TopicId) -> List[TopicSnapshotScheduler]:
		criteria = [
			EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='topic_id'), operator=EntityCriteriaOperator.EQUALS, right=topic_id)
		]
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria=criteria))

	def find_all(self, tenant_id: Optional[TenantId]) -> List[TopicSnapshotScheduler]:
		criteria = []
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria=criteria))

	def find_all_enabled(self) -> List[TopicSnapshotScheduler]:
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria=[
			EntityCriteriaExpression(left=ColumnNameLiteral(columnName='enabled'), right=True)
		]))



from datetime import datetime
from typing import List, Optional

from watchmen_meta.common import TupleNotFoundException, TupleService, TupleShaper
from watchmen_model.admin import Pipeline, PipelineStage
from watchmen_model.common import ParameterJoint, PipelineId, TenantId, TopicId
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaOperator, EntityRow, \
	EntityShaper, TooManyEntitiesFoundException
from watchmen_utilities import ArrayHelper, is_not_blank


class PipelineShaper(EntityShaper):
	@staticmethod
	def serialize_stage(stage: PipelineStage) -> dict:
		if isinstance(stage, dict):
			return stage
		else:
			return stage.dict()

	@staticmethod
	def serialize_stages(stages: Optional[List[PipelineStage]]) -> Optional[list]:
		if stages is None:
			return None
		return ArrayHelper(stages).map(lambda x: PipelineShaper.serialize_stage(x)).to_list()

	@staticmethod
	def serialize_prerequisite_on(on: Optional[ParameterJoint]) -> Optional[dict]:
		if on is None:
			return None
		return on.to_dict()

	def serialize(self, pipeline: Pipeline) -> EntityRow:
		return TupleShaper.serialize_tenant_based(pipeline, {
			'pipeline_id': pipeline.pipelineId,
			'topic_id': pipeline.topicId,
			'name': pipeline.name,
			'type': pipeline.type,
			'prerequisite_enabled': pipeline.conditional,
			'prerequisite_on': PipelineShaper.serialize_prerequisite_on(pipeline.on),
			'stages': PipelineShaper.serialize_stages(pipeline.stages),
			'enabled': pipeline.enabled,
			'validated': pipeline.validated
		})

	def deserialize(self, row: EntityRow) -> Pipeline:
		# noinspection PyTypeChecker
		return TupleShaper.deserialize_tenant_based(row, Pipeline(
			pipelineId=row.get('pipeline_id'),
			topicId=row.get('topic_id'),
			name=row.get('name'),
			type=row.get('type'),
			conditional=row.get('prerequisite_enabled'),
			on=row.get('prerequisite_on'),
			stages=row.get('stages'),
			enabled=row.get('enabled'),
			validated=row.get('validated')
		))


PIPELINE_ENTITY_NAME = 'pipelines'
PIPELINE_ENTITY_SHAPER = PipelineShaper()


class PipelineService(TupleService):
	def get_entity_name(self) -> str:
		return PIPELINE_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return PIPELINE_ENTITY_SHAPER

	def get_storable_id(self, storable: Pipeline) -> PipelineId:
		return storable.pipelineId

	def set_storable_id(self, storable: Pipeline, storable_id: PipelineId) -> Pipeline:
		storable.pipelineId = storable_id
		return storable

	def get_storable_id_column_name(self) -> str:
		return 'pipeline_id'

	# noinspection DuplicatedCode
	def find_tenant_id(self, pipeline_id: PipelineId) -> Optional[TenantId]:
		finder = self.get_entity_finder_for_columns(
			criteria=[
				EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName=self.get_storable_id_column_name()), right=pipeline_id),
			],
			distinctColumnNames=['tenant_id']
		)
		# noinspection PyTypeChecker
		rows: List[Pipeline] = self.storage.find_distinct_values(finder)
		count = len(rows)
		if count == 0:
			return None
		elif count == 1:
			return rows[0].tenantId
		else:
			raise TooManyEntitiesFoundException(f'Too many entities found by finder[{finder}].')

	def find_by_topic_id(self, topic_id: TopicId, tenant_id: TenantId) -> List[Pipeline]:
		finder = self.get_entity_finder_for_columns(
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='topic_id'), right=topic_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id),
			],
			distinctColumnNames=['tenant_id']
		)
		# noinspection PyTypeChecker
		return self.storage.find(finder)

	def update_name(self, pipeline_id: PipelineId, name: str, tenant_id: TenantId) -> Pipeline:
		"""
		update name will not increase optimistic lock version
		"""
		last_modified_at = self.now()
		last_modified_by = self.principalService.get_user_id()
		updated_count = self.storage.update_only(self.get_entity_updater(
			criteria=[
				EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName=self.get_storable_id_column_name()), right=pipeline_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
			],
			update={
				'name': name,
				'last_modified_at': last_modified_at,
				'last_modified_by': last_modified_by
			}
		))
		if updated_count == 0:
			raise TupleNotFoundException('Update 0 row might be caused by tuple not found.')
		# noinspection PyTypeChecker
		return self.find_by_id(pipeline_id)

	def update_enablement(self, pipeline_id: PipelineId, enabled: bool, tenant_id: TenantId) -> Pipeline:
		"""
		update enablement will not increase optimistic lock version
		"""
		last_modified_at = self.now()
		last_modified_by = self.principalService.get_user_id()
		updated_count = self.storage.update_only(self.get_entity_updater(
			criteria=[
				EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName=self.get_storable_id_column_name()), right=pipeline_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
			],
			update={
				'enabled': enabled,
				'last_modified_at': last_modified_at,
				'last_modified_by': last_modified_by
			}
		))
		if updated_count == 0:
			raise TupleNotFoundException('Update 0 row might be caused by tuple not found.')
		# noinspection PyTypeChecker
		return self.find_by_id(pipeline_id)

	def find_all(self, tenant_id: Optional[TenantId]) -> List[Pipeline]:
		criteria = []
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria))

	# noinspection DuplicatedCode
	def find_modified_after(self, last_modified_at: datetime, tenant_id: Optional[TenantId]) -> List[Pipeline]:
		criteria = [
			EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='last_modified_at'),
				operator=EntityCriteriaOperator.GREATER_THAN_OR_EQUALS,
				right=last_modified_at
			)
		]
		if is_not_blank(tenant_id):
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria))



from .connected_space_graphic_service import ConnectedSpaceGraphicService
from .connected_space_service import ConnectedSpaceService
from .dashboard_service import DashboardService
from .report_service import ReportService
from .subject_service import SubjectService



from typing import List, Optional

from watchmen_meta.common import TupleService, TupleShaper
from watchmen_model.common import FactorId, TenantId, TopicId
from watchmen_model.dqc import MonitorRule, MonitorRuleCode, MonitorRuleGrade, MonitorRuleId
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaOperator, EntityName, \
	EntityRow, EntityShaper
from watchmen_utilities import is_not_blank


class MonitorRuleShaper(EntityShaper):
	def serialize(self, monitor_rule: MonitorRule) -> EntityRow:
		return TupleShaper.serialize_tenant_based(monitor_rule, {
			'rule_id': monitor_rule.ruleId,
			'code': monitor_rule.code,
			'grade': monitor_rule.grade,
			'severity': monitor_rule.severity,
			'topic_id': monitor_rule.topicId,
			'factor_id': monitor_rule.factorId,
			'params': monitor_rule.params.to_dict() if monitor_rule.params is not None else None,
			'enabled': monitor_rule.enabled
		})

	def deserialize(self, row: EntityRow) -> MonitorRule:
		# noinspection PyTypeChecker
		return TupleShaper.deserialize_tenant_based(row, MonitorRule(
			ruleId=row.get('rule_id'),
			code=row.get('code'),
			grade=row.get('grade'),
			severity=row.get('severity'),
			topicId=row.get('topic_id'),
			factorId=row.get('factor_id'),
			params=row.get('params'),
			enabled=row.get('enabled')
		))


MONITOR_RULE_ENTITY_NAME = 'monitor_rules'
MONITOR_RULE_ENTITY_SHAPER = MonitorRuleShaper()


class MonitorRuleService(TupleService):
	# noinspection PyMethodMayBeStatic
	def get_entity_name(self) -> str:
		return MONITOR_RULE_ENTITY_NAME

	# noinspection PyMethodMayBeStatic
	def get_entity_shaper(self) -> EntityShaper:
		return MONITOR_RULE_ENTITY_SHAPER

	def get_storable_id_column_name(self) -> EntityName:
		return 'rule_id'

	def get_storable_id(self, storable: MonitorRule) -> MonitorRuleId:
		return storable.ruleId

	def set_storable_id(self, storable: MonitorRule, storable_id: MonitorRuleId) -> MonitorRule:
		storable.ruleId = storable_id
		return storable

	def find_by_grade_or_topic_id(
			self, grade: Optional[MonitorRuleGrade], topic_id: Optional[TopicId], tenant_id: TenantId
	) -> List[MonitorRule]:
		criteria = []
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		if grade is not None:
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='grade'), right=grade.value))
		if is_not_blank(topic_id):
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='topic_id'), right=topic_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria=criteria))

	def find_by_location(
			self, code: MonitorRuleCode, topic_id: Optional[TopicId], factor_id: Optional[FactorId],
			tenant_id: TenantId
	) -> Optional[MonitorRule]:
		criteria = []
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		if code is not None:
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='code'), right=code.value))
		if is_not_blank(topic_id):
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='topic_id'), right=topic_id))
		else:
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='topic_id'), operator=EntityCriteriaOperator.IS_EMPTY))
		if is_not_blank(factor_id):
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='factor_id'), right=factor_id))
		else:
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='factor_id'), operator=EntityCriteriaOperator.IS_EMPTY))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria=criteria))

	def delete_others_by_ids(self, rule_ids: List[MonitorRuleId], tenant_id: TenantId) -> List[MonitorRule]:
		criteria = [
			EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id),
			EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='rule_id'), operator=EntityCriteriaOperator.NOT_IN,
				right=rule_ids)
		]

		# noinspection PyTypeChecker
		return self.storage.delete_and_pull(self.get_entity_deleter(criteria=criteria))



from datetime import datetime
from typing import List, Optional

from watchmen_meta.common import AuditableShaper, LastVisitShaper, TupleNotFoundException, \
	UserBasedTupleService, UserBasedTupleShaper
from watchmen_model.common import ConnectedSpaceId, SpaceId, TenantId, UserId
from watchmen_model.console import ConnectedSpace
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaOperator, EntityRow, \
	EntityShaper


class ConnectedSpaceShaper(EntityShaper):
	def serialize(self, connected_space: ConnectedSpace) -> EntityRow:
		row = {
			'connect_id': connected_space.connectId,
			'space_id': connected_space.spaceId,
			'name': connected_space.name,
			'is_template': connected_space.isTemplate
		}
		row = AuditableShaper.serialize(connected_space, row)
		row = UserBasedTupleShaper.serialize(connected_space, row)
		row = LastVisitShaper.serialize(connected_space, row)
		return row

	def deserialize(self, row: EntityRow) -> ConnectedSpace:
		connected_space = ConnectedSpace(
			connectId=row.get('connect_id'),
			spaceId=row.get('space_id'),
			name=row.get('name'),
			isTemplate=row.get('is_template')
		)
		# noinspection PyTypeChecker
		connected_space: ConnectedSpace = AuditableShaper.deserialize(row, connected_space)
		# noinspection PyTypeChecker
		connected_space: ConnectedSpace = UserBasedTupleShaper.deserialize(row, connected_space)
		# noinspection PyTypeChecker
		connected_space: ConnectedSpace = LastVisitShaper.deserialize(row, connected_space)
		return connected_space


CONNECTED_SPACE_ENTITY_NAME = 'connected_spaces'
CONNECTED_SPACE_ENTITY_SHAPER = ConnectedSpaceShaper()


class ConnectedSpaceService(UserBasedTupleService):
	def get_entity_name(self) -> str:
		return CONNECTED_SPACE_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return CONNECTED_SPACE_ENTITY_SHAPER

	def get_storable_id_column_name(self) -> str:
		return 'connect_id'

	def get_storable_id(self, storable: ConnectedSpace) -> ConnectedSpaceId:
		return storable.connectId

	def set_storable_id(self, storable: ConnectedSpace, storable_id: ConnectedSpaceId) -> ConnectedSpace:
		storable.connectId = storable_id
		return storable

	def find_templates_by_ids(
			self, connect_ids: List[ConnectedSpaceId], space_id: Optional[SpaceId], tenant_id: TenantId
	) -> List[ConnectedSpace]:
		criteria = [
			EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName=self.get_storable_id_column_name()),
				operator=EntityCriteriaOperator.IN, right=connect_ids),
			EntityCriteriaExpression(left=ColumnNameLiteral(columnName='is_template'), right=True),
			EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
		]
		if space_id is not None and len(space_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='space_id'), right=space_id))

		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria=criteria))

	def find_templates_by_space_id(self, space_id: SpaceId, tenant_id: TenantId) -> List[ConnectedSpace]:
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='space_id'), right=space_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='is_template'), right=True),
			]
		))

	def find_templates_by_tenant_id(self, tenant_id: TenantId) -> List[ConnectedSpace]:
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='is_template'), right=True),
			]
		))

	# noinspection DuplicatedCode
	def update_name(self, connect_id: ConnectedSpaceId, name: str, user_id: UserId, tenant_id: TenantId) -> datetime:
		"""
		update name will not increase optimistic lock version
		"""
		last_modified_at = self.now()
		last_modified_by = self.principalService.get_user_id()
		updated_count = self.storage.update_only(self.get_entity_updater(
			criteria=[
				EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName=self.get_storable_id_column_name()), right=connect_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='user_id'), right=user_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
			],
			update={
				'name': name,
				'last_modified_at': last_modified_at,
				'last_modified_by': last_modified_by
			}
		))
		if updated_count == 0:
			raise TupleNotFoundException('Update 0 row might be caused by tuple not found.')
		return last_modified_at

	# noinspection DuplicatedCode
	def update_as_template(
			self, connect_id: ConnectedSpaceId, is_template: bool,
			user_id: UserId, tenant_id: TenantId) -> datetime:
		"""
		update name will not increase optimistic lock version
		"""
		last_modified_at = self.now()
		last_modified_by = self.principalService.get_user_id()
		updated_count = self.storage.update_only(self.get_entity_updater(
			criteria=[
				EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName=self.get_storable_id_column_name()), right=connect_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='user_id'), right=user_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
			],
			update={
				'is_template': is_template,
				'last_modified_at': last_modified_at,
				'last_modified_by': last_modified_by
			}
		))
		if updated_count == 0:
			raise TupleNotFoundException('Update 0 row might be caused by tuple not found.')
		return last_modified_at

	def update_last_visit_time(self, connect_id: ConnectedSpaceId) -> datetime:
		now = self.now()
		self.storage.update(self.get_entity_updater(
			criteria=[
				EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName=self.get_storable_id_column_name()), right=connect_id)
			],
			update={'last_visit_time': now}
		))
		return now



from datetime import date
from typing import Optional

from watchmen_auth import PrincipalService
from watchmen_meta.common import StorageService
from watchmen_model.admin import TopicSnapshotFrequency, TopicSnapshotJobLock, TopicSnapshotJobLockId, \
	TopicSnapshotSchedulerId
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityFinder, EntityHelper, EntityIdHelper, \
	EntityRow, \
	EntityShaper, SnowflakeGenerator, \
	TransactionalStorageSPI
from watchmen_utilities import get_current_time_in_seconds


class TopicSnapshotJobLockShaper(EntityShaper):
	def serialize(self, lock: TopicSnapshotJobLock) -> EntityRow:
		return {
			'lock_id': lock.lockId,
			'tenant_id': lock.tenantId,
			'scheduler_id': lock.schedulerId,
			'frequency': lock.frequency,
			'process_date': lock.processDate,
			'row_count': lock.rowCount,
			'status': lock.status,
			'user_id': lock.userId,
			'created_at': lock.createdAt
		}

	def deserialize(self, row: EntityRow) -> TopicSnapshotJobLock:
		return TopicSnapshotJobLock(
			lockId=row.get('lock_id'),
			tenantId=row.get('tenant_id'),
			schedulerId=row.get('scheduler_id'),
			frequency=row.get('frequency'),
			processDate=row.get('process_date'),
			rowCount=row.get('row_count'),
			status=row.get('status'),
			userId=row.get('user_id'),
			createdAt=row.get('created_at')
		)


TOPIC_SNAPSHOT_JOB_LOCK_ENTITY_NAME = 'snapshot_job_locks'
TOPIC_SNAPSHOT_JOB_LOCK_ENTITY_SHAPER = TopicSnapshotJobLockShaper()


class TopicSnapshotJobLockService(StorageService):
	def __init__(
			self,
			storage: TransactionalStorageSPI,
			snowflake_generator: SnowflakeGenerator,
			principal_service: PrincipalService
	):
		super().__init__(storage)
		self.with_snowflake_generator(snowflake_generator)
		self.with_principal_service(principal_service)

	# noinspection PyMethodMayBeStatic
	def get_entity_name(self) -> str:
		return TOPIC_SNAPSHOT_JOB_LOCK_ENTITY_NAME

	# noinspection PyMethodMayBeStatic
	def get_entity_shaper(self) -> EntityShaper:
		return TOPIC_SNAPSHOT_JOB_LOCK_ENTITY_SHAPER

	def get_entity_helper(self) -> EntityHelper:
		return EntityHelper(name=self.get_entity_name(), shaper=self.get_entity_shaper())

	def generate_lock_id(self) -> TopicSnapshotJobLockId:
		return str(self.snowflakeGenerator.next_id())

	# noinspection PyMethodMayBeStatic
	def get_lock_id_column_name(self) -> str:
		return 'lock_id'

	def get_entity_id_helper(self) -> EntityIdHelper:
		return EntityIdHelper(
			name=self.get_entity_name(), shaper=self.get_entity_shaper(),
			idColumnName=self.get_lock_id_column_name())

	def find_by_id(self, lock_id: TopicSnapshotJobLockId) -> Optional[TopicSnapshotJobLock]:
		return self.storage.find_by_id(lock_id, self.get_entity_id_helper())

	def find_by_scheduler_and_process_date(
			self, scheduler_id: TopicSnapshotSchedulerId,
			frequency: TopicSnapshotFrequency, process_date: date) -> Optional[TopicSnapshotJobLock]:
		return self.storage.find_one(EntityFinder(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='scheduler_id'), right=scheduler_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='frequency'), right=frequency),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='process_date'), right=process_date),
			]
		))

	def create(self, lock: TopicSnapshotJobLock) -> TopicSnapshotJobLock:
		lock.lockId = self.generate_lock_id()
		lock.createdAt = get_current_time_in_seconds()

		self.storage.insert_one(lock, self.get_entity_helper())
		return lock

	def update(self, lock: TopicSnapshotJobLock) -> TopicSnapshotJobLock:
		self.storage.update_one(lock, self.get_entity_id_helper())
		return lock



from typing import List, Optional

from watchmen_meta.common import TupleService, TupleShaper
from watchmen_model.admin import User, UserRole
from watchmen_model.common import DataPage, Pageable, TenantId, UserId
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaJoint, \
	EntityCriteriaJointConjunction, EntityCriteriaOperator, EntityRow, EntityShaper


class UserShaper(EntityShaper):
	def serialize(self, user: User) -> EntityRow:
		return TupleShaper.serialize_tenant_based(user, {
			'user_id': user.userId,
			'name': user.name,
			'nickname': user.nickName,
			'password': user.password,
			'is_active': user.isActive,
			'group_ids': user.groupIds,
			'role': user.role,
		})

	def deserialize(self, row: EntityRow) -> User:
		# noinspection PyTypeChecker
		return TupleShaper.deserialize_tenant_based(row, User(
			userId=row.get('user_id'),
			name=row.get('name'),
			nickName=row.get('nickname'),
			password=row.get('password'),
			isActive=row.get('is_active'),
			groupIds=row.get('group_ids'),
			role=row.get('role')
		))


USER_ENTITY_NAME = 'users'
USER_ENTITY_SHAPER = UserShaper()


class UserService(TupleService):
	def get_entity_name(self) -> str:
		return USER_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return USER_ENTITY_SHAPER

	def get_storable_id(self, storable: User) -> UserId:
		return storable.userId

	def set_storable_id(self, storable: User, storable_id: UserId) -> User:
		storable.userId = storable_id
		return storable

	def get_storable_id_column_name(self) -> str:
		return 'user_id'

	# noinspection DuplicatedCode
	def find_page_by_text(self, text: Optional[str], tenant_id: Optional[TenantId], pageable: Pageable) -> DataPage:
		# always ignore super admin
		criteria = [
			EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='role'), operator=EntityCriteriaOperator.NOT_EQUALS,
				right=UserRole.SUPER_ADMIN)
		]
		if text is not None and len(text.strip()) != 0:
			criteria.append(EntityCriteriaJoint(
				conjunction=EntityCriteriaJointConjunction.OR,
				children=[
					EntityCriteriaExpression(
						left=ColumnNameLiteral(columnName='name'), operator=EntityCriteriaOperator.LIKE, right=text),
					EntityCriteriaExpression(
						left=ColumnNameLiteral(columnName='nickname'), operator=EntityCriteriaOperator.LIKE, right=text)
				]
			))
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		return self.storage.page(self.get_entity_pager(criteria=criteria, pageable=pageable))

	# noinspection DuplicatedCode
	def find_by_name(self, text: Optional[str], tenant_id: Optional[TenantId]) -> List[User]:
		# always ignore super admin
		criteria = [
			EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='role'), operator=EntityCriteriaOperator.NOT_EQUALS,
				right=UserRole.SUPER_ADMIN)
		]
		if text is not None and len(text.strip()) != 0:
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='name'), operator=EntityCriteriaOperator.LIKE, right=text))
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria=criteria))

	def find_by_ids(self, user_ids: List[UserId], tenant_id: Optional[TenantId]) -> List[User]:
		# always ignore super admin
		criteria = [
			EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='role'), operator=EntityCriteriaOperator.NOT_EQUALS,
				right=UserRole.SUPER_ADMIN),
			EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='user_id'), operator=EntityCriteriaOperator.IN, right=user_ids)
		]
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria))

	def find_admin(self, tenant_id: TenantId) -> Optional[User]:
		criteria = [
			EntityCriteriaExpression(left=ColumnNameLiteral(columnName='role'), right=UserRole.ADMIN),
			EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
		]
		# noinspection PyTypeChecker
		admins: List[User] = self.storage.find(self.get_entity_finder(criteria=criteria))
		if len(admins) == 0:
			return None
		else:
			return admins[0]



from watchmen_model.common import LastVisit
from watchmen_storage import EntityRow


class LastVisitShaper:
	@staticmethod
	def serialize(last_visit: LastVisit, row: EntityRow) -> EntityRow:
		row['last_visit_time'] = last_visit.lastVisitTime
		return row

	@staticmethod
	def deserialize(row: EntityRow, last_visit: LastVisit) -> LastVisit:
		last_visit.lastVisitTime = row.get('last_visit_time')
		return last_visit



from typing import List, Optional

from watchmen_auth import PrincipalService
from watchmen_meta.common import StorageService
from watchmen_model.common import PatId, TenantId, UserId
from watchmen_model.system import PersonalAccessToken
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityFinder, EntityHelper, EntityIdHelper, \
	EntityRow, EntityShaper, SnowflakeGenerator, TransactionalStorageSPI
from watchmen_utilities import get_current_time_in_seconds


class PatShaper(EntityShaper):
	def serialize(self, pat: PersonalAccessToken) -> EntityRow:
		return {
			'pat_id': pat.patId,
			'token': pat.token,
			'user_id': pat.userId,
			'username': pat.username,
			'tenant_id': pat.tenantId,
			'note': pat.note,
			'expired': pat.expired,
			'permissions': pat.permissions,
			'created_at': pat.createdAt
		}

	def deserialize(self, row: EntityRow) -> PersonalAccessToken:
		return PersonalAccessToken(
			patId=row.get('pat_id'),
			token=row.get('token'),
			userId=row.get('user_id'),
			username=row.get('username'),
			tenantId=row.get('tenant_id'),
			note=row.get('note'),
			expired=row.get('expired'),
			permissions=row.get('permissions'),
			createdAt=row.get('created_at')
		)


PAT_ENTITY_NAME = 'pats'
PAT_ENTITY_SHAPER = PatShaper()


class PatService(StorageService):
	def __init__(
			self,
			storage: TransactionalStorageSPI,
			snowflake_generator: SnowflakeGenerator,
			principal_service: PrincipalService
	):
		super().__init__(storage)
		self.with_snowflake_generator(snowflake_generator)
		self.with_principal_service(principal_service)

	# noinspection PyMethodMayBeStatic
	def get_entity_name(self) -> str:
		return PAT_ENTITY_NAME

	# noinspection PyMethodMayBeStatic
	def get_entity_shaper(self) -> EntityShaper:
		return PAT_ENTITY_SHAPER

	def get_entity_helper(self) -> EntityHelper:
		return EntityHelper(name=self.get_entity_name(), shaper=self.get_entity_shaper())

	def generate_pat_id(self) -> PatId:
		return str(self.snowflakeGenerator.next_id())

	# noinspection PyMethodMayBeStatic
	def get_pat_id_column_name(self) -> str:
		return 'pat_id'

	def get_entity_id_helper(self) -> EntityIdHelper:
		return EntityIdHelper(
			name=self.get_entity_name(), shaper=self.get_entity_shaper(), idColumnName=self.get_pat_id_column_name())

	def create(self, a_pat: PersonalAccessToken) -> PersonalAccessToken:
		a_pat.patId = self.generate_pat_id()
		a_pat.createdAt = get_current_time_in_seconds()

		self.storage.insert_one(a_pat, self.get_entity_helper())
		return a_pat

	def find_by_id(self, pat_id: PatId, user_id: UserId, tenant_id: TenantId) -> Optional[PersonalAccessToken]:
		# noinspection PyTypeChecker
		pat: PersonalAccessToken = self.storage.find_one(EntityFinder(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=[
				EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName=self.get_pat_id_column_name()), right=pat_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='user_id'), right=user_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
			]
		))
		if pat is None:
			return None
		return pat

	def find_by_user_id(self, user_id: UserId, tenant_id: TenantId) -> List[PersonalAccessToken]:
		# noinspection PyTypeChecker
		return self.storage.find(EntityFinder(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='user_id'), right=user_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
			]
		))

	def delete_by_id(self, pat_id: PatId) -> None:
		self.storage.delete_by_id(pat_id, EntityIdHelper(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			idColumnName=self.get_pat_id_column_name()
		))



from typing import List

from watchmen_meta.common import TupleService, TupleShaper
from watchmen_model.common import TenantId
from watchmen_model.dqc import Catalog, CatalogCriteria, CatalogId
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaOperator, EntityRow, \
	EntityShaper
from watchmen_utilities import is_not_blank


class CatalogShaper(EntityShaper):
	def serialize(self, catalog: Catalog) -> EntityRow:
		return TupleShaper.serialize_tenant_based(catalog, {
			'catalog_id': catalog.catalogId,
			'name': catalog.name,
			'topic_ids': catalog.topicIds,
			'tech_owner_id': catalog.techOwnerId,
			'biz_owner_id': catalog.bizOwnerId,
			'tags': catalog.tags,
			'description': catalog.description
		})

	def deserialize(self, row: EntityRow) -> Catalog:
		# noinspection PyTypeChecker
		return TupleShaper.deserialize_tenant_based(row, Catalog(
			catalogId=row.get('catalog_id'),
			name=row.get('name'),
			topicIds=row.get('topic_ids'),
			techOwnerId=row.get('tech_owner_id'),
			bizOwnerId=row.get('biz_owner_id'),
			tags=row.get('tags'),
			description=row.get('description')
		))


CATALOG_ENTITY_NAME = 'catalogs'
CATALOG_ENTITY_SHAPER = CatalogShaper()


class CatalogService(TupleService):
	def get_entity_name(self) -> str:
		return CATALOG_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return CATALOG_ENTITY_SHAPER

	def get_storable_id(self, storable: Catalog) -> CatalogId:
		return storable.catalogId

	def set_storable_id(self, storable: Catalog, storable_id: CatalogId) -> Catalog:
		storable.catalogId = storable_id
		return storable

	def get_storable_id_column_name(self) -> str:
		return 'catalog_id'

	def find_by_criteria(self, criteria: CatalogCriteria, tenant_id: TenantId) -> List[Catalog]:
		storage_criteria = []
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			storage_criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		if is_not_blank(criteria.name):
			storage_criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='name'), operator=EntityCriteriaOperator.LIKE,
				right=criteria.name.strip()))
		if is_not_blank(criteria.techOwnerId):
			storage_criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='tech_owner_id'), right=criteria.techOwnerId))
		if is_not_blank(criteria.bizOwnerId):
			storage_criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='biz_owner_id'), right=criteria.bizOwnerId))
		if is_not_blank(criteria.topicId):
			storage_criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='topic_ids'), operator=EntityCriteriaOperator.LIKE,
				right=criteria.topicId))
		# noinspection PyTypeChecker

		return self.storage.find(self.get_entity_finder(criteria=storage_criteria))



from typing import List, Optional

from watchmen_meta.common import TupleService, TupleShaper
from watchmen_model.common import DataPage, ExternalWriterId, Pageable, TenantId
from watchmen_model.system import ExternalWriter
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaOperator, EntityRow, \
	EntityShaper


class ExternalWriterShaper(EntityShaper):
	def serialize(self, external_writer: ExternalWriter) -> EntityRow:
		return TupleShaper.serialize_tenant_based(external_writer, {
			'writer_id': external_writer.writerId,
			'writer_code': external_writer.writerCode,
			'type': external_writer.type,
			'pat': external_writer.pat,
			'url': external_writer.url
		})

	def deserialize(self, row: EntityRow) -> ExternalWriter:
		# noinspection PyTypeChecker
		return TupleShaper.deserialize_tenant_based(row, ExternalWriter(
			writerId=row.get('writer_id'),
			writerCode=row.get('writer_code'),
			type=row.get('type'),
			pat=row.get('pat'),
			url=row.get('url')
		))


EXTERNAL_WRITER_ENTITY_NAME = 'external_writers'
EXTERNAL_WRITER_ENTITY_SHAPER = ExternalWriterShaper()


class ExternalWriterService(TupleService):
	def get_entity_name(self) -> str:
		return EXTERNAL_WRITER_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return EXTERNAL_WRITER_ENTITY_SHAPER

	def get_storable_id(self, storable: ExternalWriter) -> ExternalWriterId:
		return storable.writerId

	def set_storable_id(self, storable: ExternalWriter, storable_id: ExternalWriterId) -> ExternalWriter:
		storable.writerId = storable_id
		return storable

	def get_storable_id_column_name(self) -> str:
		return 'writer_id'

	# noinspection DuplicatedCode
	def find_by_text(
			self, text: Optional[str], tenant_id: Optional[TenantId], pageable: Pageable) -> DataPage:
		criteria = []
		if text is not None and len(text.strip()) != 0:
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='name'), operator=EntityCriteriaOperator.LIKE, right=text))
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		return self.storage.page(self.get_entity_pager(criteria, pageable))

	def find_all(self, tenant_id: Optional[TenantId]) -> List[ExternalWriter]:
		criteria = []
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria))



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime
from typing import List, Optional, Tuple, TypeVar

from watchmen_auth import PrincipalService
from watchmen_model.common import Auditable, OptimisticLock, Pageable, Storable, UserId
from watchmen_storage import EntityColumnName, EntityCriteria, EntityDeleter, EntityDistinctValuesFinder, \
	EntityFinder, EntityHelper, EntityIdHelper, EntityName, EntityPager, EntityRow, EntityShaper, EntitySort, \
	EntityUpdate, EntityUpdater, SnowflakeGenerator, TransactionalStorageSPI
from watchmen_utilities import get_current_time_in_seconds

StorableId = TypeVar('StorableId', bound=str)
TupleId = TypeVar('TupleId', bound=str)


class TupleNotFoundException(Exception):
	pass


class StorageService(ABC):
	storage: TransactionalStorageSPI
	principalService: Optional[PrincipalService] = None
	snowflakeGenerator: Optional[SnowflakeGenerator] = None

	def __init__(self, storage: TransactionalStorageSPI):
		self.storage = storage

	def with_principal_service(self, principal_service: PrincipalService) -> StorageService:
		self.principalService = principal_service
		return self

	def with_snowflake_generator(self, snowflake_generator: SnowflakeGenerator) -> StorageService:
		self.snowflakeGenerator = snowflake_generator
		return self

	def begin_transaction(self):
		self.storage.begin()

	def commit_transaction(self):
		self.storage.commit_and_close()

	def rollback_transaction(self):
		self.storage.rollback_and_close()

	def close_transaction(self):
		self.storage.close()

	# noinspection PyMethodMayBeStatic
	def now(self) -> datetime:
		"""
		get current time in seconds
		"""
		return get_current_time_in_seconds()

	def try_to_prepare_auditable_on_create(self, storable: Storable) -> None:
		"""
		set 4 audit columns when given storable is an auditable
		"""
		if isinstance(storable, Auditable):
			now = self.now()
			storable.createdAt = now
			storable.createdBy = self.principalService.get_user_id()
			storable.lastModifiedAt = now
			storable.lastModifiedBy = self.principalService.get_user_id()

	# noinspection PyMethodMayBeStatic
	def try_to_prepare_optimistic_lock_on_create(self, storable: Storable) -> None:
		"""
		set version to 1 if given storable is an optimistic lock
		"""
		if isinstance(storable, OptimisticLock):
			storable.version = 1

	def try_to_prepare_auditable_on_update(self, storable: Storable) -> Tuple[Optional[datetime], Optional[UserId]]:
		"""
		set last modified columns when given storable is an auditable
		"""
		if isinstance(storable, Auditable):
			last_modified_at = storable.lastModifiedAt
			last_modified_by = storable.lastModifiedBy
			storable.lastModifiedAt = self.now()
			storable.lastModifiedBy = self.principalService.get_user_id()
			return last_modified_at, last_modified_by
		else:
			return None, None

	# noinspection PyMethodMayBeStatic
	def try_to_recover_auditable_on_update(
			self, storable: Storable, last_modified_at: Optional[datetime], last_modified_by: Optional[UserId]
	) -> None:
		if isinstance(storable, Auditable):
			storable.lastModifiedAt = last_modified_at
			storable.lastModifiedBy = last_modified_by

	# noinspection PyMethodMayBeStatic
	def try_to_prepare_optimistic_lock_on_update(self, storable: Storable) -> Tuple[bool, int]:
		"""
		return (true, original version) when given storable is an optimistic lock, and assign new version to given storable
		return (false, 0) when given storable is not an optimistic lock
		"""
		if isinstance(storable, OptimisticLock):
			version = storable.version
			storable.version = version + 1
			return True, version
		else:
			return False, 0

	# noinspection PyMethodMayBeStatic
	def try_to_ignore_created_columns(self, data: EntityRow) -> EntityRow:
		if 'created_at' in data:
			del data['created_at']
		if 'created_by' in data:
			del data['created_by']

		return data


class IdentifiedStorableService(StorageService):
	@abstractmethod
	def get_storable_id_column_name(self) -> EntityName:
		pass

	@abstractmethod
	def get_storable_id(self, storable: Storable) -> StorableId:
		pass

	@abstractmethod
	def set_storable_id(self, storable: Storable, storable_id: StorableId) -> Storable:
		"""
		return exactly the given storable
		"""
		pass

	@staticmethod
	def is_storable_id_faked(storable_id: StorableId) -> bool:
		if storable_id is None:
			return True

		trimmed_storable_id = storable_id.strip()
		if len(trimmed_storable_id) == 0:
			return True
		elif trimmed_storable_id.startswith('f-'):
			return True
		else:
			return False

	def generate_storable_id(self) -> StorableId:
		return str(self.snowflakeGenerator.next_id())

	def redress_storable_id(self, storable: Storable) -> Storable:
		"""
		return exactly the given tuple, replace by generated id if it is faked
		"""
		if IdentifiedStorableService.is_storable_id_faked(self.get_storable_id(storable)):
			self.set_storable_id(storable, self.generate_storable_id())
		return storable

	# noinspection PyMethodMayBeStatic
	def get_optimistic_column_name(self) -> str:
		return 'version'

	def ignore_optimistic_keys(self, data: EntityRow) -> EntityRow:
		if self.get_optimistic_column_name() in data:
			del data[self.get_optimistic_column_name()]
		if self.get_storable_id_column_name() in data:
			del data[self.get_storable_id_column_name()]

		return data

	def ignore_storable_id(self, data: EntityRow) -> EntityRow:
		if self.get_storable_id_column_name() in data:
			del data[self.get_storable_id_column_name()]
		return data


class EntityService(IdentifiedStorableService):
	@abstractmethod
	def get_entity_name(self) -> EntityName:
		pass

	@abstractmethod
	def get_entity_shaper(self) -> EntityShaper:
		pass

	def get_entity_helper(self) -> EntityHelper:
		return EntityHelper(name=self.get_entity_name(), shaper=self.get_entity_shaper())

	def get_entity_id_helper(self) -> EntityIdHelper:
		return EntityIdHelper(
			name=self.get_entity_name(), shaper=self.get_entity_shaper(),
			idColumnName=self.get_storable_id_column_name())

	def get_entity_finder(self, criteria: EntityCriteria, sort: Optional[EntitySort] = None) -> EntityFinder:
		return EntityFinder(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=criteria,
			sort=sort
		)

	def get_entity_finder_for_columns(
			self, criteria: EntityCriteria,
			distinctColumnNames: List[EntityColumnName],
			distinctValueOnSingleColumn: Optional[bool] = False,
			sort: Optional[EntitySort] = None
	) -> EntityDistinctValuesFinder:
		return EntityDistinctValuesFinder(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=criteria,
			sort=sort,
			distinctColumnNames=distinctColumnNames,
			distinctValueOnSingleColumn=distinctValueOnSingleColumn
		)

	def get_entity_pager(
			self, criteria: EntityCriteria, pageable: Pageable, sort: Optional[EntitySort] = None
	) -> EntityPager:
		return EntityPager(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=criteria,
			sort=sort,
			pageable=pageable
		)

	def get_entity_updater(self, criteria: EntityCriteria, update: EntityUpdate) -> EntityUpdater:
		return EntityUpdater(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=criteria,
			update=update
		)

	def get_entity_deleter(self, criteria: EntityCriteria) -> EntityDeleter:
		return EntityDeleter(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=criteria
		)



from typing import Optional

from watchmen_auth import PrincipalService
from watchmen_meta.common import LastVisitShaper, StorageService, UserBasedTupleShaper
from watchmen_model.common import TenantId, UserId
from watchmen_model.gui import LastSnapshot
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityFinder, EntityHelper, EntityIdHelper, \
	EntityRow, EntityShaper, EntityUpdater, TransactionalStorageSPI


class LastSnapshotShaper(EntityShaper):
	def serialize(self, last_snapshot: LastSnapshot) -> EntityRow:
		row = {
			'language': last_snapshot.language,
			'last_dashboard_id': last_snapshot.lastDashboardId,
			'admin_dashboard_id': last_snapshot.adminDashboardId,
			'favorite_pin': last_snapshot.favoritePin
		}
		row = UserBasedTupleShaper.serialize(last_snapshot, row)
		row = LastVisitShaper.serialize(last_snapshot, row)
		return row

	def deserialize(self, row: EntityRow) -> LastSnapshot:
		last_snapshot = LastSnapshot(
			language=row.get('language'),
			lastDashboardId=row.get('last_dashboard_id'),
			adminDashboardId=row.get('admin_dashboard_id'),
			favoritePin=row.get('favorite_pin')
		)
		# noinspection PyTypeChecker
		last_snapshot: LastSnapshot = UserBasedTupleShaper.deserialize(row, last_snapshot)
		# noinspection PyTypeChecker
		last_snapshot: LastSnapshot = LastVisitShaper.deserialize(row, last_snapshot)
		return last_snapshot


LAST_SNAPSHOT_ENTITY_NAME = 'last_snapshots'
LAST_SNAPSHOT_ENTITY_SHAPER = LastSnapshotShaper()


class LastSnapshotService(StorageService):
	def __init__(self, storage: TransactionalStorageSPI, principal_service: PrincipalService):
		super().__init__(storage)
		self.with_principal_service(principal_service)

	# noinspection PyMethodMayBeStatic
	def get_entity_name(self) -> str:
		return LAST_SNAPSHOT_ENTITY_NAME

	# noinspection PyMethodMayBeStatic
	def get_entity_shaper(self) -> EntityShaper:
		return LAST_SNAPSHOT_ENTITY_SHAPER

	def get_entity_helper(self) -> EntityHelper:
		return EntityHelper(name=self.get_entity_name(), shaper=self.get_entity_shaper())

	def get_entity_id_helper(self) -> EntityIdHelper:
		return EntityIdHelper(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			idColumnName='user_id'
		)

	def create(self, last_snapshot: LastSnapshot) -> LastSnapshot:
		self.storage.insert_one(last_snapshot, self.get_entity_helper())
		return last_snapshot

	def update(self, last_snapshot: LastSnapshot) -> LastSnapshot:
		self.storage.update_only(EntityUpdater(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='user_id'), right=last_snapshot.userId),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=last_snapshot.tenantId)
			],
			update={
				'language': last_snapshot.language,
				'last_dashboard_id': last_snapshot.lastDashboardId,
				'admin_dashboard_id': last_snapshot.adminDashboardId,
				'favorite_pin': last_snapshot.favoritePin,
				'last_visit_time': last_snapshot.lastVisitTime
			}
		))
		return last_snapshot

	def find_by_user_id(self, user_id: UserId, tenant_id: TenantId) -> Optional[LastSnapshot]:
		return self.storage.find_one(EntityFinder(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='user_id'), right=user_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
			]
		))

	def delete_by_id(self, user_id: UserId) -> Optional[LastSnapshot]:
		return self.storage.delete_by_id_and_pull(user_id, self.get_entity_id_helper())



from datetime import datetime
from typing import Optional, Union

from watchmen_meta.common import AuditableShaper, LastVisitShaper, TupleNotFoundException, \
	UserBasedTupleService, UserBasedTupleShaper
from watchmen_model.common import DashboardId, TenantId, UserId
from watchmen_model.console import Dashboard, DashboardParagraph, DashboardReport
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityRow, EntityShaper
from watchmen_utilities import ArrayHelper


class DashboardShaper(EntityShaper):
	# noinspection PyMethodMayBeStatic
	def serialize_to_dict(
			self,
			data: Optional[Union[dict, DashboardReport, DashboardParagraph]]
	) -> Optional[dict]:
		if data is None:
			return None
		elif isinstance(data, dict):
			return data
		else:
			return data.dict()

	def serialize(self, dashboard: Dashboard) -> EntityRow:
		row = {
			'dashboard_id': dashboard.dashboardId,
			'name': dashboard.name,
			'reports': ArrayHelper(dashboard.reports).map(lambda x: self.serialize_to_dict(x)).to_list(),
			'paragraphs': ArrayHelper(dashboard.paragraphs).map(lambda x: self.serialize_to_dict(x)).to_list(),
			'auto_refresh_interval': dashboard.autoRefreshInterval
		}
		row = AuditableShaper.serialize(dashboard, row)
		row = UserBasedTupleShaper.serialize(dashboard, row)
		row = LastVisitShaper.serialize(dashboard, row)
		return row

	def deserialize(self, row: EntityRow) -> Dashboard:
		dashboard = Dashboard(
			dashboardId=row.get('dashboard_id'),
			name=row.get('name'),
			reports=row.get('reports'),
			paragraphs=row.get('paragraphs'),
			autoRefreshInterval=row.get('auto_refresh_interval')
		)
		# noinspection PyTypeChecker
		dashboard: Dashboard = AuditableShaper.deserialize(row, dashboard)
		# noinspection PyTypeChecker
		dashboard: Dashboard = UserBasedTupleShaper.deserialize(row, dashboard)
		# noinspection PyTypeChecker
		dashboard: Dashboard = LastVisitShaper.deserialize(row, dashboard)
		return dashboard


DASHBOARD_ENTITY_NAME = 'dashboards'
DASHBOARD_ENTITY_SHAPER = DashboardShaper()


class DashboardService(UserBasedTupleService):
	def get_entity_name(self) -> str:
		return DASHBOARD_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return DASHBOARD_ENTITY_SHAPER

	def get_storable_id_column_name(self) -> str:
		return 'dashboard_id'

	def get_storable_id(self, storable: Dashboard) -> DashboardId:
		return storable.dashboardId

	def set_storable_id(self, storable: Dashboard, storable_id: DashboardId) -> Dashboard:
		storable.dashboardId = storable_id
		return storable

	# noinspection DuplicatedCode
	def update_name(self, dashboard_id: DashboardId, name: str, user_id: UserId, tenant_id: TenantId) -> datetime:
		"""
		update name will not increase optimistic lock version
		"""
		last_modified_at = self.now()
		last_modified_by = self.principalService.get_user_id()
		updated_count = self.storage.update_only(self.get_entity_updater(
			criteria=[
				EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName=self.get_storable_id_column_name()), right=dashboard_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='user_id'), right=user_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
			],
			update={
				'name': name,
				'last_modified_at': last_modified_at,
				'last_modified_by': last_modified_by
			}
		))
		if updated_count == 0:
			raise TupleNotFoundException('Update 0 row might be caused by tuple not found.')
		return last_modified_at



from typing import List, Optional, Tuple

from watchmen_auth import PrincipalService
from watchmen_model.common import Auditable, OptimisticLock, TenantId, UserBasedTuple, UserId
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityRow, SnowflakeGenerator, \
	TooManyEntitiesFoundException, TransactionalStorageSPI
from .storage_service import EntityService, TupleId, TupleNotFoundException
from .tuple_service import AuditableShaper, OptimisticLockShaper


class UserBasedTupleShaper:
	@staticmethod
	def serialize(user_based_tuple: UserBasedTuple, row: EntityRow) -> EntityRow:
		row['user_id'] = user_based_tuple.userId
		row['tenant_id'] = user_based_tuple.tenantId
		if isinstance(user_based_tuple, Auditable):
			row = AuditableShaper.serialize(user_based_tuple, row)
		if isinstance(user_based_tuple, OptimisticLock):
			row = OptimisticLockShaper.serialize(user_based_tuple, row)
		return row

	@staticmethod
	def deserialize(row: EntityRow, user_based_tuple: UserBasedTuple) -> UserBasedTuple:
		user_based_tuple.userId = row.get('user_id')
		user_based_tuple.tenantId = row.get('tenant_id')
		if isinstance(user_based_tuple, Auditable):
			user_based_tuple = AuditableShaper.deserialize(row, user_based_tuple)
		if isinstance(user_based_tuple, OptimisticLock):
			user_based_tuple = OptimisticLockShaper.deserialize(row, user_based_tuple)
		return user_based_tuple


# noinspection PyAbstractClass
class UserBasedTupleService(EntityService):
	def __init__(
			self,
			storage: TransactionalStorageSPI,
			snowflake_generator: SnowflakeGenerator, principal_service: PrincipalService
	):
		super().__init__(storage)
		self.with_snowflake_generator(snowflake_generator)
		self.with_principal_service(principal_service)

	def create(self, a_tuple: UserBasedTuple) -> UserBasedTuple:
		self.try_to_prepare_auditable_on_create(a_tuple)
		self.storage.insert_one(a_tuple, self.get_entity_helper())
		return a_tuple

	def update(self, a_tuple: UserBasedTuple) -> UserBasedTuple:
		original_last_modified_at, original_last_modified_by = self.try_to_prepare_auditable_on_update(a_tuple)
		if isinstance(a_tuple, Auditable):
			updated_count = self.storage.update_only(self.get_entity_updater(
				criteria=[
					EntityCriteriaExpression(
						left=ColumnNameLiteral(columnName=self.get_storable_id_column_name()),
						right=self.get_storable_id(a_tuple))
				],
				# to avoid update created columns in update
				update=self.try_to_ignore_created_columns(
					self.ignore_storable_id(self.get_entity_shaper().serialize(a_tuple)))
			))
		else:
			updated_count = self.storage.update_one(a_tuple, self.get_entity_id_helper())
		if updated_count == 0:
			self.try_to_recover_auditable_on_update(a_tuple, original_last_modified_at, original_last_modified_by)
			raise TupleNotFoundException('Update 0 row might be caused by tuple not found.')
		return a_tuple

	def delete(self, tuple_id: TupleId) -> Optional[UserBasedTuple]:
		return self.storage.delete_by_id_and_pull(tuple_id, self.get_entity_id_helper())

	def find_by_id(self, tuple_id: TupleId) -> Optional[UserBasedTuple]:
		return self.storage.find_one(self.get_entity_finder(
			criteria=[
				EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName=self.get_storable_id_column_name()), right=tuple_id),
			]
		))

	def find_all_by_user_id(self, user_id: UserId, tenant_id: TenantId) -> List[UserBasedTuple]:
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='user_id'), right=user_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
			]
		))

	def find_tenant_and_user(self, tuple_id: TupleId) -> Optional[Tuple[TenantId, UserId]]:
		finder = self.get_entity_finder_for_columns(
			criteria=[
				EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName=self.get_storable_id_column_name()), right=tuple_id),
			],
			distinctColumnNames=['tenant_id', 'user_id']
		)
		# noinspection PyTypeChecker
		rows: List[UserBasedTuple] = self.storage.find_distinct_values(finder)
		count = len(rows)
		if count == 0:
			return None
		elif count == 1:
			return rows[0].tenantId, rows[0].userId
		else:
			raise TooManyEntitiesFoundException(f'Too many entities found by finder[{finder}].')

	def delete_by_id(self, tuple_id: TupleId) -> None:
		self.storage.delete_by_id(tuple_id, self.get_entity_id_helper())



from datetime import datetime
from typing import Callable, Optional

from watchmen_meta.admin.user_service import USER_ENTITY_NAME, USER_ENTITY_SHAPER
from watchmen_meta.common import ask_meta_storage
from watchmen_meta.system.pat_service import PAT_ENTITY_NAME, PAT_ENTITY_SHAPER
from watchmen_model.admin import User
from watchmen_model.system import PersonalAccessToken
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityFinder, EntityIdHelper, \
	TransactionalStorageSPI


def redress_user(user: Optional[User], clear_pwd: bool) -> Optional[User]:
	if user is None:
		return None
	if not user.isActive:
		return None
	if clear_pwd:
		del user.password
	return user


def find_user_by_name(storage: TransactionalStorageSPI, username: str, clear_pwd: bool) -> Optional[User]:
	if username is None or len(username.strip()) == 0:
		return None

	storage.begin()
	try:
		# noinspection PyTypeChecker
		user: User = storage.find_one(EntityFinder(
			name=USER_ENTITY_NAME,
			shaper=USER_ENTITY_SHAPER,
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='name'), right=username)
			]
		))
		return redress_user(user, clear_pwd)
	finally:
		storage.close()


def build_find_user_by_name(clear_pwd: bool = True) -> Callable[[str], Optional[User]]:
	"""
	autonomous transaction inside
	"""
	return lambda username: find_user_by_name(ask_meta_storage(), username, clear_pwd)


def find_pat_by_token(storage: TransactionalStorageSPI, pat_token: str) -> Optional[PersonalAccessToken]:
	if pat_token is None or len(pat_token.strip()) == 0:
		return None
	# noinspection PyTypeChecker
	pat: PersonalAccessToken = storage.find_one(EntityFinder(
		name=PAT_ENTITY_NAME,
		shaper=PAT_ENTITY_SHAPER,
		criteria=[
			EntityCriteriaExpression(left=ColumnNameLiteral(columnName='token'), right=pat_token)
		]
	))
	if pat is None:
		return None
	if pat.expired is None:
		# no expired date
		return pat
	if pat.expired < datetime.now().replace(tzinfo=None):
		# pat is expired
		return None
	return pat


def find_user_by_pat(storage: TransactionalStorageSPI, pat_token: str) -> Optional[User]:
	"""
	find unexpired pat, find active user. Otherwise, return none
	"""

	storage.begin()
	try:
		pat = find_pat_by_token(storage, pat_token)
		if pat is None:
			return None
		# noinspection PyTypeChecker
		user: User = storage.find_by_id(pat.userId, EntityIdHelper(
			name=USER_ENTITY_NAME,
			shaper=USER_ENTITY_SHAPER,
			idColumnName='user_id'
		))
		return redress_user(user, True)
	finally:
		storage.close()


def build_find_user_by_pat() -> Callable[[str], Optional[User]]:
	"""
	autonomous transaction inside
	"""
	return lambda pat_token: find_user_by_pat(ask_meta_storage(), pat_token)



from watchmen_meta.common import ask_engine_index_enabled, StorageService
from watchmen_model.admin import Pipeline
from watchmen_model.analysis import PipelineIndex
from watchmen_model.common import PipelineId
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityDeleter, EntityRow, EntityShaper, \
	SnowflakeGenerator, TransactionalStorageSPI


class PipelineIndexShaper(EntityShaper):
	def serialize(self, pipeline_index: PipelineIndex) -> EntityRow:
		return {
			'pipeline_index_id': pipeline_index.pipelineIndexId,
			'pipeline_id': pipeline_index.pipelineId,
			'pipeline_name': pipeline_index.pipelineName,
			'stage_id': pipeline_index.stageId,
			'stage_name': pipeline_index.stageName,
			'unit_id': pipeline_index.unitId,
			'unit_name': pipeline_index.unitName,
			'action_id': pipeline_index.actionId,
			'mapping_to_topic_id': pipeline_index.mappingToTopicId,
			'mapping_to_factor_id': pipeline_index.mappingToFactorId,
			'source_from_topic_id': pipeline_index.sourceFromTopicId,
			'source_from_factor_id': pipeline_index.sourceFromFactorId,
			'ref_type': pipeline_index.refType,
			'tenant_id': pipeline_index.tenantId,
			'created_at': pipeline_index.createdAt,
			'last_modified_at': pipeline_index.lastModifiedAt
		}

	def deserialize(self, row: EntityRow) -> PipelineIndex:
		return PipelineIndex(
			pipelineIndexId=row.get('pipeline_index_id'),
			pipelineId=row.get('pipeline_id'),
			pipelineName=row.get('pipeline_name'),
			stageId=row.get('stage_id'),
			stageName=row.get('stage_name'),
			unitId=row.get('unit_id'),
			unitName=row.get('unit_name'),
			actionId=row.get('action_id'),
			mappingToTopicId=row.get('mapping_to_topic_id'),
			mappingToFactorId=row.get('mapping_to_factor_id'),
			sourceFromTopicId=row.get('source_from_topic_id'),
			sourceFromFactorId=row.get('source_from_factor_id'),
			refType=row.get('ref_type'),
			tenantId=row.get('tenant_id'),
			createdAt=row.get('created_at'),
			lastModifiedAt=row.get('last_modified_at')
		)


PIPELINE_INDEX_ENTITY_NAME = 'pipeline_index'
PIPELINE_INDEX_ENTITY_SHAPER = PipelineIndexShaper()


class PipelineIndexService(StorageService):
	def __init__(
			self, storage: TransactionalStorageSPI, snowflake_generator: SnowflakeGenerator
	):
		super().__init__(storage)
		self.with_snowflake_generator(snowflake_generator)

	# noinspection PyMethodMayBeStatic,PyUnusedLocal
	def build_index(self, pipeline: Pipeline) -> None:
		if not ask_engine_index_enabled():
			return
		# TODO build pipeline index
		return

	def update_index_on_name_changed(self, pipeline: Pipeline) -> None:
		pass

	def update_index_on_enablement_changed(self, pipeline: Pipeline) -> None:
		pass

	def remove_index(self, pipeline_id: PipelineId) -> None:
		if not ask_engine_index_enabled():
			return

		self.storage.delete(EntityDeleter(
			name=PIPELINE_INDEX_ENTITY_NAME,
			shaper=PIPELINE_INDEX_ENTITY_SHAPER,
			criteria=[EntityCriteriaExpression(left=ColumnNameLiteral(columnName='pipeline_id'), right=pipeline_id)]
		))



from typing import Optional

from watchmen_auth import PrincipalService
from watchmen_meta.common import LastVisitShaper, StorageService, UserBasedTupleShaper
from watchmen_model.common import TenantId, UserId
from watchmen_model.gui import Favorite
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityFinder, EntityHelper, EntityIdHelper, \
	EntityRow, EntityShaper, EntityUpdater, TransactionalStorageSPI


class FavoriteShaper(EntityShaper):
	def serialize(self, favorite: Favorite) -> EntityRow:
		row = {
			'connected_space_ids': favorite.connectedSpaceIds,
			'dashboard_ids': favorite.dashboardIds
		}
		row = UserBasedTupleShaper.serialize(favorite, row)
		row = LastVisitShaper.serialize(favorite, row)
		return row

	def deserialize(self, row: EntityRow) -> Favorite:
		favorite = Favorite(
			connectedSpaceIds=row.get('connected_space_ids'),
			dashboardIds=row.get('dashboard_ids')
		)
		# noinspection PyTypeChecker
		favorite: Favorite = UserBasedTupleShaper.deserialize(row, favorite)
		# noinspection PyTypeChecker
		favorite: Favorite = LastVisitShaper.deserialize(row, favorite)
		return favorite


FAVORITE_ENTITY_NAME = 'favorites'
FAVORITE_ENTITY_SHAPER = FavoriteShaper()


class FavoriteService(StorageService):
	def __init__(self, storage: TransactionalStorageSPI, principal_service: PrincipalService):
		super().__init__(storage)
		self.with_principal_service(principal_service)

	# noinspection PyMethodMayBeStatic
	def get_entity_name(self) -> str:
		return FAVORITE_ENTITY_NAME

	# noinspection PyMethodMayBeStatic
	def get_entity_shaper(self) -> EntityShaper:
		return FAVORITE_ENTITY_SHAPER

	def get_entity_helper(self) -> EntityHelper:
		return EntityHelper(name=self.get_entity_name(), shaper=self.get_entity_shaper())

	def get_entity_id_helper(self) -> EntityIdHelper:
		return EntityIdHelper(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			idColumnName='user_id'
		)

	def create(self, favorite: Favorite) -> Favorite:
		self.storage.insert_one(favorite, self.get_entity_helper())
		return favorite

	def update(self, favorite: Favorite) -> Favorite:
		self.storage.update_only(EntityUpdater(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='user_id'), right=favorite.userId),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=favorite.tenantId)
			],
			update={
				'connected_space_ids': favorite.connectedSpaceIds,
				'dashboard_ids': favorite.dashboardIds,
				'last_visit_time': favorite.lastVisitTime
			}
		))
		return favorite

	def find_by_user_id(self, user_id: UserId, tenant_id: TenantId) -> Optional[Favorite]:
		return self.storage.find_one(EntityFinder(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='user_id'), right=user_id),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
			]
		))

	def delete_by_id(self, user_id: UserId) -> Optional[Favorite]:
		return self.storage.delete_by_id_and_pull(user_id, self.get_entity_id_helper())



from typing import List, Optional

from watchmen_meta.common import TupleService, TupleShaper
from watchmen_model.common import DataPage, Pageable, TenantId
from watchmen_model.system import Tenant
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaOperator, EntityRow, \
	EntityShaper


class TenantShaper(EntityShaper):
	def serialize(self, tenant: Tenant) -> EntityRow:
		return TupleShaper.serialize(tenant, {
			'tenant_id': tenant.tenantId,
			'name': tenant.name
		})

	def deserialize(self, row: EntityRow) -> Tenant:
		# noinspection PyTypeChecker
		return TupleShaper.deserialize(row, Tenant(
			tenantId=row.get('tenant_id'),
			name=row.get('name')
		))


TENANT_ENTITY_NAME = 'tenants'
TENANT_ENTITY_SHAPER = TenantShaper()


class TenantService(TupleService):
	def get_entity_name(self) -> str:
		return TENANT_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return TENANT_ENTITY_SHAPER

	def get_storable_id(self, storable: Tenant) -> TenantId:
		return storable.tenantId

	def set_storable_id(self, storable: Tenant, storable_id: TenantId) -> Tenant:
		storable.tenantId = storable_id
		return storable

	def get_storable_id_column_name(self) -> str:
		return 'tenant_id'

	def find_by_text(self, text: Optional[str], pageable: Pageable) -> DataPage:
		criteria = []
		if text is not None and len(text.strip()) != 0:
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='name'), operator=EntityCriteriaOperator.LIKE, right=text))
		return self.storage.page(self.get_entity_pager(criteria, pageable))

	def find_all(self) -> List[Tenant]:
		# noinspection PyTypeChecker
		return self.storage.find_all(self.get_entity_helper())



from typing import Optional

from watchmen_meta.common import StorageService
from watchmen_model.common import TenantId
from watchmen_model.system import KeyStore
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityFinder, EntityRow, EntityShaper


class KeyStoreShaper(EntityShaper):
	def serialize(self, key_store: KeyStore) -> EntityRow:
		return {
			'tenant_id': key_store.tenantId,
			'key_type': key_store.keyType,
			'params': key_store.params,
			'created_at': key_store.createdAt,
			'created_by': key_store.createdBy
		}

	def deserialize(self, row: EntityRow) -> KeyStore:
		return KeyStore(
			tenantId=row.get('tenant_id'),
			keyType=row.get('key_type'),
			params=row.get('params'),
			createdAt=row.get('created_at'),
			createdBy=row.get('created_by')
		)


KEY_STORE_ENTITY_NAME = 'key_stores'
KEY_STORE_ENTITY_SHAPER = KeyStoreShaper()


class KeyStoreService(StorageService):
	# noinspection PyMethodMayBeStatic
	def get_entity_name(self) -> str:
		return KEY_STORE_ENTITY_NAME

	# noinspection PyMethodMayBeStatic
	def get_entity_shaper(self) -> EntityShaper:
		return KEY_STORE_ENTITY_SHAPER

	def find_by_type(self, key_type: str, tenant_id: TenantId) -> Optional[KeyStore]:
		return self.storage.find_one(EntityFinder(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=[
				EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName='key_type'), right=key_type),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
			]
		))



from logging import getLogger
from typing import Callable, Optional

from pydantic import BaseSettings

from watchmen_auth import PrincipalService
from watchmen_model.admin import User, UserRole
from watchmen_model.common import TenantId, UserId
from watchmen_model.system import DataSourceType
from watchmen_storage import competitive_worker_id, CompetitiveWorkerRestarter, CompetitiveWorkerShutdownSignal, \
	immutable_worker_id, SnowflakeGenerator, StorageBasedWorkerIdGenerator, TransactionalStorageSPI
from .exception import InitialMetaAppException

logger = getLogger(__name__)


class MetaSettings(BaseSettings):
	SUPER_ADMIN_TENANT_ID: TenantId = '1'
	SUPER_ADMIN_USER_ID: UserId = '1'
	SUPER_ADMIN_USER_NAME: str = 'imma-super'
	SUPER_ADMIN_USER_NICKNAME: str = 'IMMA Super'

	META_STORAGE_TYPE: DataSourceType = DataSourceType.MYSQL
	META_STORAGE_USER_NAME: str = 'watchmen'
	META_STORAGE_PASSWORD: str = 'watchmen'
	META_STORAGE_HOST: str = 'localhost'
	META_STORAGE_PORT: int = 3306
	META_STORAGE_NAME: str = 'watchmen'
	META_STORAGE_ECHO: bool = False

	SNOWFLAKE_DATA_CENTER_ID: int = 0  # data center id
	SNOWFLAKE_WORKER_ID: int = 0  # worker id
	SNOWFLAKE_COMPETITIVE_WORKERS: bool = True  # enable competitive snowflake worker
	SNOWFLAKE_COMPETITIVE_WORKER_HEART_BEAT_INTERVAL: int = 60  # competitive worker heart beat interval, in seconds
	SNOWFLAKE_COMPETITIVE_WORKER_CREATION_RETRY_TIMES: int = 3  # competitive worker creation max retry times
	SNOWFLAKE_COMPETITIVE_WORKER_RESTART_ON_SHOWDOWN: bool = False  # competitive worker restart automatically on shutdown

	ENGINE_INDEX: bool = True

	class Config:
		# secrets_dir = '/var/run'
		env_file = '.env'
		env_file_encoding = 'utf-8'
		case_sensitive = True


settings = MetaSettings()
logger.info(f'Meta settings[{settings.dict()}].')


def ask_super_admin() -> PrincipalService:
	return PrincipalService(User(
		userId=settings.SUPER_ADMIN_USER_ID,
		name=settings.SUPER_ADMIN_USER_NAME,
		nickname=settings.SUPER_ADMIN_USER_NICKNAME,
		isActive=True,
		groupIds=[],
		tenantId=settings.SUPER_ADMIN_TENANT_ID,
		role=UserRole.SUPER_ADMIN
	))


# noinspection DuplicatedCode
def build_mysql_storage() -> Callable[[], TransactionalStorageSPI]:
	from watchmen_storage_mysql import StorageMySQLConfiguration
	configuration = StorageMySQLConfiguration.config() \
		.host(settings.META_STORAGE_HOST, settings.META_STORAGE_PORT) \
		.account(settings.META_STORAGE_USER_NAME, settings.META_STORAGE_PASSWORD) \
		.schema(settings.META_STORAGE_NAME) \
		.echo(settings.META_STORAGE_ECHO)
	return lambda: configuration.build()


def build_oracle_storage() -> Callable[[], TransactionalStorageSPI]:
	from watchmen_storage_oracle import StorageOracleConfiguration
	configuration = StorageOracleConfiguration.config() \
		.host(settings.META_STORAGE_HOST, settings.META_STORAGE_PORT) \
		.account(settings.META_STORAGE_USER_NAME, settings.META_STORAGE_PASSWORD) \
		.schema(settings.META_STORAGE_NAME) \
		.echo(settings.META_STORAGE_ECHO)
	return lambda: configuration.build()


# noinspection DuplicatedCode
def build_mongodb_storage() -> Callable[[], TransactionalStorageSPI]:
	from watchmen_storage_mongodb import StorageMongoConfiguration
	configuration = StorageMongoConfiguration.config() \
		.host(settings.META_STORAGE_HOST, settings.META_STORAGE_PORT) \
		.account(settings.META_STORAGE_USER_NAME, settings.META_STORAGE_PASSWORD) \
		.schema(settings.META_STORAGE_NAME) \
		.echo(settings.META_STORAGE_ECHO)
	return lambda: configuration.build()


def build_mssql_storage() -> Callable[[], TransactionalStorageSPI]:
	from watchmen_storage_mssql import StorageMSSQLConfiguration
	configuration = StorageMSSQLConfiguration.config() \
		.host(settings.META_STORAGE_HOST, settings.META_STORAGE_PORT) \
		.account(settings.META_STORAGE_USER_NAME, settings.META_STORAGE_PASSWORD) \
		.schema(settings.META_STORAGE_NAME) \
		.echo(settings.META_STORAGE_ECHO)
	return lambda: configuration.build()


def build_postgresql_storage() -> Callable[[], TransactionalStorageSPI]:
	from watchmen_storage_postgresql import StoragePostgreSQLConfiguration
	configuration = StoragePostgreSQLConfiguration.config() \
		.host(settings.META_STORAGE_HOST, settings.META_STORAGE_PORT) \
		.account(settings.META_STORAGE_USER_NAME, settings.META_STORAGE_PASSWORD) \
		.schema(settings.META_STORAGE_NAME) \
		.echo(settings.META_STORAGE_ECHO)
	return lambda: configuration.build()


class MetaStorageHolder:
	metaStorage: Optional[Callable[[], TransactionalStorageSPI]] = None


meta_storage_holder = MetaStorageHolder()


def build_meta_storage() -> Callable[[], TransactionalStorageSPI]:
	storage_type = settings.META_STORAGE_TYPE
	if storage_type == DataSourceType.MYSQL:
		return build_mysql_storage()
	if storage_type == DataSourceType.ORACLE:
		return build_oracle_storage()
	if storage_type == DataSourceType.MONGODB:
		return build_mongodb_storage()
	if storage_type == DataSourceType.MSSQL:
		return build_mssql_storage()
	if storage_type == DataSourceType.POSTGRESQL:
		return build_postgresql_storage()

	raise InitialMetaAppException(f'Meta storage type[{storage_type}] is not supported yet.')


def ask_meta_storage() -> TransactionalStorageSPI:
	"""
	build a new meta storage instance
	"""
	if meta_storage_holder.metaStorage is None:
		meta_storage_holder.metaStorage = build_meta_storage()
	return meta_storage_holder.metaStorage()


class SnowflakeGeneratorHolder:
	snowflakeGenerator: Optional[SnowflakeGenerator] = None


def build_snowflake_generator(storage: TransactionalStorageSPI) -> SnowflakeGenerator:
	if settings.SNOWFLAKE_COMPETITIVE_WORKERS:
		# competitive workers
		def shutdown_listener(
				signal: CompetitiveWorkerShutdownSignal,
				data_center_id: int, worker_id: int,
				restart: CompetitiveWorkerRestarter) -> None:
			logger.warning(
				f'Worker[dataCenterId={data_center_id}, workerId={worker_id}] shutdown on signal[{signal}].')
			if settings.SNOWFLAKE_COMPETITIVE_WORKER_RESTART_ON_SHOWDOWN:
				if signal == CompetitiveWorkerShutdownSignal.EXCEPTION_RAISED:
					restart()

		worker_id_generator = competitive_worker_id(StorageBasedWorkerIdGenerator(
			storage=storage,
			heart_beat_interval=settings.SNOWFLAKE_COMPETITIVE_WORKER_HEART_BEAT_INTERVAL,
			worker_creation_retry_times=settings.SNOWFLAKE_COMPETITIVE_WORKER_CREATION_RETRY_TIMES,
			shutdown_listener=shutdown_listener
		))
		return SnowflakeGenerator(
			data_center_id=settings.SNOWFLAKE_DATA_CENTER_ID,
			generate_worker_id=worker_id_generator
		)
	else:
		# fix worker id
		return SnowflakeGenerator(
			data_center_id=settings.SNOWFLAKE_DATA_CENTER_ID,
			generate_worker_id=immutable_worker_id(settings.SNOWFLAKE_WORKER_ID)
		)


snowflake_generator_holder = SnowflakeGeneratorHolder()
snowflake_generator_holder.snowflakeGenerator = build_snowflake_generator(ask_meta_storage())


def ask_snowflake_generator() -> SnowflakeGenerator:
	if snowflake_generator_holder.snowflakeGenerator is None:
		snowflake_generator_holder.snowflakeGenerator = build_snowflake_generator(ask_meta_storage())
	return snowflake_generator_holder.snowflakeGenerator


def ask_engine_index_enabled() -> bool:
	return settings.ENGINE_INDEX



from watchmen_auth import PrincipalService
from watchmen_meta.common import StorageService
from watchmen_model.dqc import MonitorJobLock, MonitorJobLockId
from watchmen_storage import EntityHelper, EntityIdHelper, EntityRow, EntityShaper, SnowflakeGenerator, \
	TransactionalStorageSPI
from watchmen_utilities import get_current_time_in_seconds


class MonitorJobLockShaper(EntityShaper):
	def serialize(self, lock: MonitorJobLock) -> EntityRow:
		return {
			'lock_id': lock.lockId,
			'tenant_id': lock.tenantId,
			'topic_id': lock.topicId,
			'frequency': lock.frequency,
			'process_date': lock.processDate,
			'status': lock.status,
			'user_id': lock.userId,
			'created_at': lock.createdAt
		}

	def deserialize(self, row: EntityRow) -> MonitorJobLock:
		return MonitorJobLock(
			lockId=row.get('lock_id'),
			tenantId=row.get('tenant_id'),
			topicId=row.get('topic_id'),
			frequency=row.get('frequency'),
			processDate=row.get('process_date'),
			status=row.get('status'),
			userId=row.get('user_id'),
			createdAt=row.get('created_at')
		)


MONITOR_JOB_LOCK_ENTITY_NAME = 'monitor_job_locks'
MONITOR_JOB_LOCK_ENTITY_SHAPER = MonitorJobLockShaper()


# noinspection DuplicatedCode
class MonitorJobLockService(StorageService):
	def __init__(
			self,
			storage: TransactionalStorageSPI,
			snowflake_generator: SnowflakeGenerator,
			principal_service: PrincipalService
	):
		super().__init__(storage)
		self.with_snowflake_generator(snowflake_generator)
		self.with_principal_service(principal_service)

	# noinspection PyMethodMayBeStatic
	def get_entity_name(self) -> str:
		return MONITOR_JOB_LOCK_ENTITY_NAME

	# noinspection PyMethodMayBeStatic
	def get_entity_shaper(self) -> EntityShaper:
		return MONITOR_JOB_LOCK_ENTITY_SHAPER

	def get_entity_helper(self) -> EntityHelper:
		return EntityHelper(name=self.get_entity_name(), shaper=self.get_entity_shaper())

	def generate_lock_id(self) -> MonitorJobLockId:
		return str(self.snowflakeGenerator.next_id())

	# noinspection PyMethodMayBeStatic
	def get_lock_id_column_name(self) -> str:
		return 'lock_id'

	def get_entity_id_helper(self) -> EntityIdHelper:
		return EntityIdHelper(
			name=self.get_entity_name(), shaper=self.get_entity_shaper(),
			idColumnName=self.get_lock_id_column_name())

	def create(self, lock: MonitorJobLock) -> MonitorJobLock:
		lock.lockId = self.generate_lock_id()
		lock.createdAt = get_current_time_in_seconds()

		self.storage.insert_one(lock, self.get_entity_helper())
		return lock



from .exception import InitialMetaAppException
from .last_visit_service import LastVisitShaper
from .settings import ask_engine_index_enabled, ask_meta_storage, ask_snowflake_generator, ask_super_admin, MetaSettings
from .storage_service import EntityService, IdentifiedStorableService, StorageService, TupleNotFoundException
from .tuple_service import AuditableShaper, OptimisticLockShaper, TupleService, TupleShaper
from .user_based_tuple_service import UserBasedTupleService, UserBasedTupleShaper



from datetime import datetime
from typing import List, Optional

from watchmen_meta.common import TupleService, TupleShaper
from watchmen_model.admin import Factor, Topic, TopicType
from watchmen_model.common import DataPage, FactorId, Pageable, TenantId, TopicId
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaJoint, \
	EntityCriteriaJointConjunction, EntityCriteriaOperator, EntityDistinctValuesFinder, EntityRow, EntityShaper, \
	SnowflakeGenerator
from watchmen_utilities import ArrayHelper, is_not_blank


class TopicShaper(EntityShaper):
	def serialize(self, topic: Topic) -> EntityRow:
		return TupleShaper.serialize_tenant_based(topic, {
			'topic_id': topic.topicId,
			'name': topic.name,
			'description': topic.description,
			'type': topic.type,
			'kind': topic.kind,
			'data_source_id': topic.dataSourceId,
			'factors': ArrayHelper(topic.factors).map(lambda x: x.dict()).to_list(),
		})

	def deserialize(self, row: EntityRow) -> Topic:
		# noinspection PyTypeChecker
		return TupleShaper.deserialize_tenant_based(row, Topic(
			topicId=row.get('topic_id'),
			name=row.get('name'),
			description=row.get('description'),
			type=row.get('type'),
			kind=row.get('kind'),
			dataSourceId=row.get('data_source_id'),
			factors=row.get('factors')
		))


TOPIC_ENTITY_NAME = 'topics'
TOPIC_ENTITY_SHAPER = TopicShaper()


class FactorService:
	def __init__(self, snowflake_generator: SnowflakeGenerator):
		self.snowflakeGenerator = snowflake_generator

	def generate_factor_id(self) -> FactorId:
		return str(self.snowflakeGenerator.next_id())

	# noinspection PyMethodMayBeStatic
	def redress_factor_id(self, factor: Factor) -> Factor:
		"""
		return exactly the given factor
		"""
		# if TupleService.is_storable_id_faked(factor.factorId):
		# 	factor.factorId = self.generate_factor_id()
		return factor


class TopicService(TupleService):
	def get_entity_name(self) -> str:
		return TOPIC_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return TOPIC_ENTITY_SHAPER

	def get_storable_id(self, storable: Topic) -> TopicId:
		return storable.topicId

	def set_storable_id(self, storable: Topic, storable_id: TopicId) -> Topic:
		storable.topicId = storable_id
		return storable

	def get_storable_id_column_name(self) -> str:
		return 'topic_id'

	def find_ids_by_ids(self, topic_ids: List[TopicId], tenant_id: TenantId) -> List[TopicId]:
		"""
		find topic ids by given ids, returned list might be less than given
		"""
		# noinspection PyTypeChecker
		topics: List[Topic] = self.storage.find_distinct_values(EntityDistinctValuesFinder(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=[
				EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName='topic_id'), operator=EntityCriteriaOperator.IN, right=topic_ids),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
			],
			distinctColumnNames=['topic_id']
		))
		return ArrayHelper(topics).map(lambda x: x.topicId).to_list()

	# noinspection DuplicatedCode
	def find_page_by_text(self, text: Optional[str], tenant_id: Optional[TenantId], pageable: Pageable) -> DataPage:
		criteria = []
		if text is not None and len(text.strip()) != 0:
			criteria.append(EntityCriteriaJoint(
				conjunction=EntityCriteriaJointConjunction.OR,
				children=[
					EntityCriteriaExpression(
						left=ColumnNameLiteral(columnName='name'), operator=EntityCriteriaOperator.LIKE, right=text),
					EntityCriteriaExpression(
						left=ColumnNameLiteral(columnName='description'), operator=EntityCriteriaOperator.LIKE,
						right=text)
				]
			))
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		return self.storage.page(self.get_entity_pager(criteria=criteria, pageable=pageable))

	# noinspection DuplicatedCode
	def find_by_name(
			self, name: Optional[str], exclude_types: Optional[List[TopicType]],
			tenant_id: Optional[TenantId]) -> List[Topic]:
		criteria = []
		if name is not None and len(name.strip()) != 0:
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='name'), operator=EntityCriteriaOperator.LIKE, right=name))
		if exclude_types is not None and len(exclude_types) != 0:
			criteria.append(
				EntityCriteriaExpression(
					left=ColumnNameLiteral(columnName='type'), operator=EntityCriteriaOperator.NOT_IN,
					right=exclude_types))
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria=criteria))

	def find_by_name_and_tenant(self, name: str, tenant_id: TenantId) -> Optional[Topic]:
		return self.storage.find_one(self.get_entity_finder(
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='name'), right=name),
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id)
			]
		))

	def find_by_ids(self, topic_ids: List[TopicId], tenant_id: Optional[TenantId]) -> List[Topic]:
		criteria = [
			EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='topic_id'), operator=EntityCriteriaOperator.IN, right=topic_ids)
		]
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria))

	def find_all(self, tenant_id: Optional[TenantId]) -> List[Topic]:
		criteria = []
		if is_not_blank(tenant_id):
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria))

	# noinspection DuplicatedCode
	def find_modified_after(self, last_modified_at: datetime, tenant_id: Optional[TenantId]) -> List[Topic]:
		criteria = [
			EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='last_modified_at'),
				operator=EntityCriteriaOperator.GREATER_THAN_OR_EQUALS,
				right=last_modified_at
			)
		]
		if is_not_blank(tenant_id):
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria))



from datetime import datetime
from typing import Dict, List

from watchmen_meta.common import ask_engine_index_enabled, StorageService
from watchmen_model.admin import Factor, Topic
from watchmen_model.analysis import FactorIndex
from watchmen_model.common import FactorId, TenantId, TopicId
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaJoint, \
	EntityCriteriaJointConjunction, EntityCriteriaOperator, EntityDeleter, EntityFinder, EntityIdHelper, EntityRow, \
	EntityShaper, SnowflakeGenerator, TransactionalStorageSPI
from watchmen_utilities import ArrayHelper, get_current_time_in_seconds


class FactorIndexShaper(EntityShaper):
	def serialize(self, factor_index: FactorIndex) -> EntityRow:
		return {
			'factor_index_id': factor_index.factorIndexId,
			'factor_id': factor_index.factorId,
			'factor_type': factor_index.factorType,
			'factor_name': factor_index.factorName,
			'factor_label': factor_index.factorLabel,
			'factor_description': factor_index.factorDescription,
			'topic_id': factor_index.topicId,
			'topic_name': factor_index.topicName,
			'tenant_id': factor_index.tenantId,
			'created_at': factor_index.createdAt,
			'last_modified_at': factor_index.lastModifiedAt
		}

	def deserialize(self, row: EntityRow) -> FactorIndex:
		return FactorIndex(
			factorIndexId=row.get('factor_index_id'),
			factorId=row.get('factor_id'),
			factorType=row.get('factor_type'),
			factorName=row.get('factor_name'),
			factorLabel=row.get('factor_label'),
			factorDescription=row.get('factor_description'),
			topicId=row.get('topic_id'),
			topicName=row.get('topic_name'),
			tenantId=row.get('tenant_id'),
			createdAt=row.get('created_at'),
			lastModifiedAt=row.get('last_modified_at')
		)


FACTOR_INDEX_ENTITY_NAME = 'factor_index'
FACTOR_INDEX_ENTITY_SHAPER = FactorIndexShaper()


class TopicIndexService(StorageService):
	def __init__(
			self, storage: TransactionalStorageSPI, snowflake_generator: SnowflakeGenerator
	):
		super().__init__(storage)
		self.with_snowflake_generator(snowflake_generator)

	def build_factor_index(self, factor: Factor, topic: Topic, now: datetime) -> FactorIndex:
		return FactorIndex(
			factorIndexId=str(self.snowflakeGenerator.next_id()),
			factorId=factor.factorId,
			factorType=factor.type,
			factorName=factor.name,
			factorLabel=factor.label,
			factorDescription=factor.description,
			topicId=topic.topicId,
			topicName=topic.name,
			tenantId=topic.tenantId,
			createdAt=now,
			lastModifiedAt=now
		)

	def build_index(self, topic: Topic) -> None:
		if not ask_engine_index_enabled():
			return

		# noinspection PyTypeChecker
		factor_index_list: List[FactorIndex] = self.storage.find(EntityFinder(
			name=FACTOR_INDEX_ENTITY_NAME,
			shaper=FACTOR_INDEX_ENTITY_SHAPER,
			criteria=[EntityCriteriaExpression(left=ColumnNameLiteral(columnName='topic_id'), right=topic.topicId)]
		))
		index_map: Dict[FactorId, FactorIndex] = ArrayHelper(factor_index_list) \
			.to_map(lambda x: x.factorId, lambda x: x)
		now = get_current_time_in_seconds()
		current_index_list = ArrayHelper(topic.factors).map(lambda x: self.build_factor_index(x, topic, now)).to_list()
		entity_id_helper = EntityIdHelper(
			name=FACTOR_INDEX_ENTITY_NAME,
			shaper=FACTOR_INDEX_ENTITY_SHAPER,
			idColumnName='factor_index_id'
		)
		for factor_index in current_index_list:
			if factor_index.factorId in index_map:
				old_factor_index = index_map[factor_index.factorId]
				factor_index.factorIndexId = old_factor_index.factorIndexId
				factor_index.createdAt = old_factor_index.createdAt
				self.storage.update_one(factor_index, entity_id_helper)
			else:
				self.storage.insert_one(factor_index, entity_id_helper)
		current_factor_ids = ArrayHelper(current_index_list).map(lambda x: x.factorId).to_list()
		to_remove_list: List[FactorIndex] = ArrayHelper(list(index_map.values())) \
			.filter(lambda x: x.factorId not in current_factor_ids).to_list()
		if len(to_remove_list) != 0:
			self.storage.delete(EntityDeleter(
				name=FACTOR_INDEX_ENTITY_NAME,
				shaper=FACTOR_INDEX_ENTITY_SHAPER,
				criteria=[
					EntityCriteriaExpression(left=ColumnNameLiteral(columnName='topic_id'), right=topic.topicId),
					EntityCriteriaExpression(
						left=ColumnNameLiteral(columnName='factor_index_id'),
						operator=EntityCriteriaOperator.IN,
						right=ArrayHelper(to_remove_list).map(lambda x: x.factorIndexId).to_list()
					)
				]
			))

	def remove_index(self, topic_id: TopicId) -> None:
		if not ask_engine_index_enabled():
			return

		self.storage.delete(EntityDeleter(
			name=FACTOR_INDEX_ENTITY_NAME,
			shaper=FACTOR_INDEX_ENTITY_SHAPER,
			criteria=[EntityCriteriaExpression(left=ColumnNameLiteral(columnName='topic_id'), right=topic_id)]
		))

	def find(self, name: str, tenant_id: TenantId) -> List[FactorIndex]:
		if not ask_engine_index_enabled():
			return []
		# noinspection PyTypeChecker
		return self.storage.find(EntityFinder(
			name=FACTOR_INDEX_ENTITY_NAME,
			shaper=FACTOR_INDEX_ENTITY_SHAPER,
			criteria=[
				EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id),
				EntityCriteriaJoint(
					conjunction=EntityCriteriaJointConjunction.AND,
					children=[
						EntityCriteriaExpression(
							left=ColumnNameLiteral(columnName='factor_name'),
							operator=EntityCriteriaOperator.LIKE,
							right=name),
						EntityCriteriaExpression(
							left=ColumnNameLiteral(columnName='topic_name'),
							operator=EntityCriteriaOperator.LIKE,
							right=name)
					]
				)
			]
		))



from .enumeration_service import EnumItemService, EnumService
from .pipeline_graphic_service import PipelineGraphicService
from .pipeline_service import PipelineService
from .space_service import SpaceService
from .topic_service import FactorService, TopicService
from .topic_snapshot_lock_service import TopicSnapshotJobLockService
from .topic_snapshot_scheduler_service import TopicSnapshotSchedulerService
from .user_group_service import UserGroupService
from .user_service import UserService



from typing import List, Optional

from watchmen_meta.common import TupleService, TupleShaper
from watchmen_model.admin import Enum, EnumItem
from watchmen_model.common import DataPage, EnumId, EnumItemId, Pageable, TenantId
from watchmen_storage import ColumnNameLiteral, EntityCriteria, EntityCriteriaExpression, EntityCriteriaJoint, \
	EntityCriteriaJointConjunction, EntityCriteriaOperator, EntityDeleter, EntityFinder, EntityHelper, EntityRow, \
	EntityShaper, EntitySort, SnowflakeGenerator, TransactionalStorageSPI


class EnumShaper(EntityShaper):
	def serialize(self, an_enum: Enum) -> EntityRow:
		return TupleShaper.serialize_tenant_based(an_enum, {
			'enum_id': an_enum.enumId,
			'name': an_enum.name,
			'description': an_enum.description,
			'parent_enum_id': an_enum.parentEnumId
		})

	def deserialize(self, row: EntityRow) -> Enum:
		# noinspection PyTypeChecker
		return TupleShaper.deserialize_tenant_based(row, Enum(
			enumId=row.get('enum_id'),
			name=row.get('name'),
			description=row.get('description'),
			parentEnumId=row.get('parent_enum_id')
		))


ENUM_ENTITY_NAME = 'enums'
ENUM_ENTITY_SHAPER = EnumShaper()


class EnumService(TupleService):
	def get_entity_name(self) -> str:
		return ENUM_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return ENUM_ENTITY_SHAPER

	def get_storable_id(self, storable: Enum) -> EnumId:
		return storable.enumId

	def set_storable_id(self, storable: Enum, storable_id: EnumId) -> Enum:
		storable.enumId = storable_id
		return storable

	def get_storable_id_column_name(self) -> str:
		return 'enum_id'

	def find_by_name(self, name: str, tenant_id: Optional[TenantId]) -> Optional[Enum]:
		criteria = [EntityCriteriaExpression(left=ColumnNameLiteral(columnName='name'), right=name)]
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		return self.storage.find_one(self.get_entity_finder(criteria))

	# noinspection DuplicatedCode
	def find_by_text(self, text: Optional[str], tenant_id: Optional[TenantId], pageable: Pageable) -> DataPage:
		criteria = []
		if text is not None and len(text.strip()) != 0:
			criteria.append(EntityCriteriaJoint(
				conjunction=EntityCriteriaJointConjunction.OR,
				children=[
					EntityCriteriaExpression(
						left=ColumnNameLiteral(columnName='name'), operator=EntityCriteriaOperator.LIKE, right=text),
					EntityCriteriaExpression(
						left=ColumnNameLiteral(columnName='description'), operator=EntityCriteriaOperator.LIKE,
						right=text)
				]
			))
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		return self.storage.page(self.get_entity_pager(criteria, pageable))

	def find_all(self, tenant_id: Optional[TenantId]) -> List[Enum]:
		criteria = []
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria))


class EnumItemShaper(EntityShaper):
	def serialize(self, enum_item: EnumItem) -> EntityRow:
		return {
			'item_id': enum_item.itemId,
			'code': enum_item.code,
			'label': enum_item.label,
			'parent_code': enum_item.parentCode,
			'replace_code': enum_item.replaceCode,
			'enum_id': enum_item.enumId,
			'tenant_id': enum_item.tenantId
		}

	def deserialize(self, row: EntityRow) -> EnumItem:
		# noinspection PyTypeChecker
		return EnumItem(
			itemId=row.get('item_id'),
			code=row.get('code'),
			label=row.get('label'),
			parentCode=row.get('parent_code'),
			replaceCode=row.get('replace_code'),
			enumId=row.get('enum_id'),
			tenantId=row.get('tenant_id')
		)


ENUM_ITEM_ENTITY_NAME = 'enum_items'
ENUM_ITEM_ENTITY_SHAPER = EnumItemShaper()


class EnumItemService:
	storage: TransactionalStorageSPI

	def __init__(self, storage: TransactionalStorageSPI, snowflake_generator: SnowflakeGenerator):
		self.storage = storage
		self.snowflakeGenerator = snowflake_generator

	# noinspection PyMethodMayBeStatic
	def get_entity_name(self) -> str:
		return ENUM_ITEM_ENTITY_NAME

	# noinspection PyMethodMayBeStatic
	def get_entity_shaper(self) -> EntityShaper:
		return ENUM_ITEM_ENTITY_SHAPER

	def get_entity_helper(self) -> EntityHelper:
		return EntityHelper(name=self.get_entity_name(), shaper=self.get_entity_shaper())

	def get_entity_finder(self, criteria: EntityCriteria, sort: Optional[EntitySort] = None) -> EntityFinder:
		return EntityFinder(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=criteria,
			sort=sort
		)

	def get_entity_deleter(self, criteria: EntityCriteria) -> EntityDeleter:
		return EntityDeleter(
			name=self.get_entity_name(),
			shaper=self.get_entity_shaper(),
			criteria=criteria
		)

	def generate_item_id(self) -> EnumItemId:
		return str(self.snowflakeGenerator.next_id())

	def redress_item_id(self, enum_item: EnumItem) -> EnumItem:
		"""
		return exactly the given tuple, replace by generated id if it is faked
		"""
		if TupleService.is_storable_id_faked(enum_item.itemId):
			enum_item.itemId = self.generate_item_id()
		return enum_item

	def create(self, enum_item: EnumItem) -> EnumItem:
		self.storage.insert_one(enum_item, self.get_entity_helper())
		return enum_item

	def find_by_enum_id(self, enum_id: EnumId) -> List[EnumItem]:
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder([
			EntityCriteriaExpression(left=ColumnNameLiteral(columnName='enum_id'), right=enum_id)
		]))

	def delete_by_enum_id(self, enum_id: EnumId) -> None:
		self.storage.delete(self.get_entity_deleter([
			EntityCriteriaExpression(left=ColumnNameLiteral(columnName='enum_id'), right=enum_id)
		]))



from typing import List, Optional

from watchmen_meta.common import TupleService, TupleShaper
from watchmen_model.admin import UserGroup
from watchmen_model.common import DataPage, Pageable, TenantId, UserGroupId
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaJoint, \
	EntityCriteriaJointConjunction, EntityCriteriaOperator, EntityRow, EntityShaper


class UserGroupShaper(EntityShaper):
	def serialize(self, user_group: UserGroup) -> EntityRow:
		return TupleShaper.serialize_tenant_based(user_group, {
			'user_group_id': user_group.userGroupId,
			'name': user_group.name,
			'description': user_group.description,
			'user_ids': user_group.userIds,
			'space_ids': user_group.spaceIds,
			'indicator_ids': user_group.indicatorIds
		})

	def deserialize(self, row: EntityRow) -> UserGroup:
		# noinspection PyTypeChecker
		return TupleShaper.deserialize_tenant_based(row, UserGroup(
			userGroupId=row.get('user_group_id'),
			name=row.get('name'),
			description=row.get('description'),
			userIds=row.get('user_ids'),
			spaceIds=row.get('space_ids'),
			indicatorIds=row.get('indicator_ids')
		))


USER_GROUP_ENTITY_NAME = 'user_groups'
USER_GROUP_ENTITY_SHAPER = UserGroupShaper()


class UserGroupService(TupleService):
	def get_entity_name(self) -> str:
		return USER_GROUP_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return USER_GROUP_ENTITY_SHAPER

	def get_storable_id(self, storable: UserGroup) -> UserGroupId:
		return storable.userGroupId

	def set_storable_id(self, storable: UserGroup, storable_id: UserGroupId) -> UserGroup:
		storable.userGroupId = storable_id
		return storable

	def get_storable_id_column_name(self) -> str:
		return 'user_group_id'

	# noinspection DuplicatedCode
	def find_page_by_text(self, text: Optional[str], tenant_id: Optional[TenantId], pageable: Pageable) -> DataPage:
		criteria = []
		if text is not None and len(text.strip()) != 0:
			criteria.append(EntityCriteriaJoint(
				conjunction=EntityCriteriaJointConjunction.OR,
				children=[
					EntityCriteriaExpression(
						left=ColumnNameLiteral(columnName='name'), operator=EntityCriteriaOperator.LIKE, right=text),
					EntityCriteriaExpression(
						left=ColumnNameLiteral(columnName='description'), operator=EntityCriteriaOperator.LIKE,
						right=text)
				]
			))
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		return self.storage.page(self.get_entity_pager(criteria=criteria, pageable=pageable))

	# noinspection DuplicatedCode
	def find_by_name(self, text: Optional[str], tenant_id: Optional[TenantId]) -> List[UserGroup]:
		criteria = []
		if text is not None and len(text.strip()) != 0:
			criteria.append(EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='name'), operator=EntityCriteriaOperator.LIKE, right=text))
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria=criteria))

	def find_by_ids(self, user_group_ids: List[UserGroupId], tenant_id: Optional[TenantId]) -> List[UserGroup]:
		criteria = [
			EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='user_group_id'), operator=EntityCriteriaOperator.IN,
				right=user_group_ids)
		]
		if tenant_id is not None and len(tenant_id.strip()) != 0:
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria))



from datetime import datetime
from typing import List, Optional, Union

from watchmen_meta.common import UserBasedTupleService, UserBasedTupleShaper
from watchmen_model.admin import PipelineGraphic, TopicGraphic
from watchmen_model.common import PipelineGraphicId, TenantId, UserId
from watchmen_storage import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaOperator, \
	EntityRow, EntityShaper
from watchmen_utilities import ArrayHelper, is_not_blank


class PipelineGraphicShaper(EntityShaper):
	# noinspection PyMethodMayBeStatic
	def serialize_to_dict(
			self,
			data: Optional[Union[dict, TopicGraphic]]
	) -> Optional[dict]:
		if data is None:
			return None
		elif isinstance(data, dict):
			return data
		else:
			return data.dict()

	def serialize(self, pipeline_graphic: PipelineGraphic) -> EntityRow:
		row = {
			'pipeline_graphic_id': pipeline_graphic.pipelineGraphId,
			'name': pipeline_graphic.name,
			'topics': ArrayHelper(pipeline_graphic.topics).map(lambda x: self.serialize_to_dict(x)).to_list(),
			'created_at': pipeline_graphic.createdAt,
			'last_modified_at': pipeline_graphic.lastModifiedAt
		}
		row = UserBasedTupleShaper.serialize(pipeline_graphic, row)
		return row

	def deserialize(self, row: EntityRow) -> PipelineGraphic:
		pipeline_graphic = PipelineGraphic(
			pipelineGraphId=row.get('pipeline_graphic_id'),
			name=row.get('name'),
			topics=row.get('topics'),
			createdAt=row.get('created_at'),
			lastModifiedAt=row.get('last_modified_at')
		)
		# noinspection PyTypeChecker
		pipeline_graphic: PipelineGraphic = UserBasedTupleShaper.deserialize(row, pipeline_graphic)
		return pipeline_graphic


PIPELINE_GRAPHIC_ENTITY_NAME = 'pipeline_graphics'
PIPELINE_GRAPHIC_ENTITY_SHAPER = PipelineGraphicShaper()


class PipelineGraphicService(UserBasedTupleService):
	def get_entity_name(self) -> str:
		return PIPELINE_GRAPHIC_ENTITY_NAME

	def get_entity_shaper(self) -> EntityShaper:
		return PIPELINE_GRAPHIC_ENTITY_SHAPER

	def get_storable_id_column_name(self) -> str:
		return 'pipeline_graphic_id'

	def get_storable_id(self, storable: PipelineGraphic) -> PipelineGraphicId:
		return storable.pipelineGraphId

	def set_storable_id(self, storable: PipelineGraphic, storable_id: PipelineGraphicId) -> PipelineGraphic:
		storable.pipelineGraphId = storable_id
		return storable

	# noinspection DuplicatedCode
	def find_modified_after(
			self, last_modified_at: datetime, user_id: Optional[UserId], tenant_id: Optional[TenantId]
	) -> List[PipelineGraphic]:
		criteria = [
			EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='last_modified_at'),
				operator=EntityCriteriaOperator.GREATER_THAN_OR_EQUALS,
				right=last_modified_at
			)
		]
		if is_not_blank(user_id):
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='user_id'), right=user_id))
		if is_not_blank(tenant_id):
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find(self.get_entity_finder(criteria))

	def find_by_ids(
			self, graphic_ids: List[PipelineGraphicId], user_id: Optional[UserId], tenant_id: Optional[TenantId]
	) -> List[PipelineGraphic]:
		criteria = [
			EntityCriteriaExpression(
				left=ColumnNameLiteral(columnName='pipeline_graphic_id'),
				operator=EntityCriteriaOperator.IN,
				right=graphic_ids
			)
		]
		if is_not_blank(user_id):
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='user_id'), right=user_id))
		if is_not_blank(tenant_id):
			criteria.append(EntityCriteriaExpression(left=ColumnNameLiteral(columnName='tenant_id'), right=tenant_id))
		# noinspection PyTypeChecker
		return self.storage.find_distinct_values(self.get_entity_finder_for_columns(
			distinctColumnNames=['pipeline_graphic_id'],
			criteria=criteria
		))
